from pathlib import Path
from unittest.mock import Mock

import pytest
from bs4 import BeautifulSoup
from fasthtml import common as fh
from starlette.testclient import TestClient

from fast_gov_uk.design_system import AbstractField
from fast_gov_uk.forms import Form


@pytest.fixture
def fast():
    from .app import fast
    # Mock notify for testing
    fast.notify_client = Mock()
    return fast


@pytest.fixture
def client(fast):
    return fh.Client(fast)


@pytest.fixture
def error_client(fast):
    return TestClient(fast, raise_server_exceptions=False)


@pytest.fixture(scope="function")
def db(fast):
    db = fast.db
    db.q("BEGIN TRANSACTION;")
    yield db
    db.q("ROLLBACK;")


@pytest.fixture
def picture():
    this_file = Path(__file__).resolve()
    parent = this_file.parent
    return open(parent / "picture.png", "rb")


@pytest.fixture
def html():
    def pretty_html(x):
        if isinstance(x, AbstractField):
            html_str = fh.to_xml(x)
        elif isinstance(x, Form):
            html_str = fh.to_xml(x)
        elif isinstance(x, fh.FT):
            html_str = str(x)
        else:
            html_str = str(x)
        soup = BeautifulSoup(html_str, "html.parser")
        return soup.prettify()
    return pretty_html


@pytest.fixture
def find():
    def _find(html, tag, selector):
        soup = BeautifulSoup(html, "html.parser")
        return soup.find(tag, selector)
    return _find



"""
GOV.UK Design System components for page elements such as Header, Footer and PhaseBanner.

Includes the generic :py:meth:`Page` component that renders a standard GOV.UK Page wth
a GOV.UK Header, a GOV.UK PhaseBanner and a GOV.UK Footer and of course - whatever content
you might want to add to it.

Also includes a generic :py:meth:`Cookies` component that renders a standard GOV.UK Cookies
Page that you can customise for your service.
"""

import fasthtml.common as fh

from .components import Table
from .typography import H1, H2, P
from .utils import OGL, Crown, Logo


def Header(title: str = "", homepage: str = "/", **kwargs) -> fh.FT:
    """
    `GOV.UK Header`_ component that supports a title (name of your service)
    that is also a link to your homepage.

    If you are trying to render a standard GOV.UK page, you should probably
    use the :py:meth:`Page` component - which includes a header.

    Example:

        >>> ds.Header()
        # The basic GOV.UK header without a title or a link
        >>> ds.Header("MyService")
        # GOV.UK header with the title "MyService" that links to the page "/"
        >>> ds.Header("MyService", "/foo")
        # GOV.UK header with the title "MyService" that links to the page "/foo"

    Args:
        title (str, optional): The title of the header. Defaults to "".
        homepage (str, optional): The URL of the homepage. Defaults to "/".
        **kwargs: Additional keyword arguments.

    Returns:
        FT: A FastHTML Header component.

    .. _GOV.UK Header: https://design-system.service.gov.uk/components/header/
    """
    return fh.Header(
        fh.Div(
            fh.Div(
                fh.A(
                    fh.NotStr(Logo()),
                    fh.Span(title, cls="govuk-header__product-name"),
                    href=homepage,
                    cls="govuk-header__link govuk-header__link--homepage",
                    aria_label="Home",
                ),
                cls="govuk-header__logo",
            ),
            cls="govuk-header__container govuk-width-container",
        ),
        cls="govuk-header",
        data_module="govuk-header",
        **kwargs,
    )


def FooterLink(
    text: str,
    href: str,
    **kwargs,
) -> fh.FT:
    """
    A link in the `GOV.UK Footer`_ component.

    Unless you are trying to build your own footer, you should probably use
    the :py:meth:`Footer` component.

    Examples:

        >>> ds.FooterLink("Home", "/")
        # Renders a footer link to the home page

    Args:
        text (str): The text to display in the link.
        href (str): The URL the link points to.
        **kwargs: Additional keyword arguments.

    Returns:
        FT: A FastHTML FooterLink component.

    .. _GOV.UK Footer: https://design-system.service.gov.uk/components/footer/
    """
    return fh.A(
        text,
        href=href,
        cls="govuk-footer__link",
        **kwargs,
    )


def Footer(*links: tuple[str, str], **kwargs) -> fh.FT:
    """
    `GOV.UK Footer`_ component with optional links.

    If you are trying to render a standard GOV.UK page, you should probably
    use the :py:meth:`Page` component.

    The :py:meth:`Page` component pulls in the `Footer` from the `/footer`
    endpoint so that there is one and only one place with the definition for
    the Footer of your service.

    Examples:

        >>> ds.Footer(("Home", "/"), ("Feedback", "/feedback"))
        # Renders a GOV.UK Footer with links to the home page and the feedback page

    Args:
        *links (tuple[str, str]): Footer links as (text, href).
        **kwargs: Additional keyword arguments.

    Returns:
        FT: A FastHTML Footer component.

    .. _GOV.UK Footer: https://design-system.service.gov.uk/components/footer/
    """
    return fh.Footer(
        fh.Div(
            fh.NotStr(Crown()),
            fh.Div(
                fh.Div(
                    fh.H2("Support links", cls="govuk-visually-hidden")
                    if links
                    else "",
                    fh.Ul(
                        *[
                            fh.Li(
                                FooterLink(text, href),
                                cls="govuk-footer__inline-list-item",
                            )
                            for text, href in links
                        ],
                        cls="govuk-footer__inline-list",
                    )
                    if links
                    else "",
                    fh.NotStr(OGL()),
                    fh.Span(
                        "All content is available under the",
                        fh.A(
                            "Open Government Licence v3.0",
                            cls="govuk-footer__link",
                            href="https://www.nationalarchives.gov.uk/doc/open-government-licence/version/3/",
                            rel="license",
                        ),
                        cls="govuk-footer__licence-description",
                    ),
                    cls="govuk-footer__meta-item govuk-footer__meta-item--grow",
                ),
                fh.Div(
                    fh.A(
                        "© Crown copyright",
                        href="https://www.nationalarchives.gov.uk/information-management/re-using-public-sector-information/uk-government-licensing-framework/crown-copyright/",
                        cls="govuk-footer__link govuk-footer__copyright-logo",
                    ),
                    cls="govuk-footer__meta-item",
                ),
                cls="govuk-footer__meta",
            ),
            cls="govuk-width-container",
        ),
        cls="govuk-footer",
        **kwargs,
    )


def PhaseBanner(
    *content: fh.FT,
    phase: str = "Alpha",
    **kwargs,
) -> fh.FT:
    """
    A `GOV.UK Phase Banner`_ component - specifying if your service is e.g. "Alpha" or "Beta".

    If you are trying to render a standard GOV.UK page, you should probably
    use the :py:meth:`Page` component.

    The :py:meth:`Page` component pulls in the `PhaseBanner` from the `/phase`
    endpoint so that there is one and only one place with the definition for
    the PhaseBanner of your service.

    Examples:

        >>> ds.PhaseBanner("This is a new service")
        # Renders a GOV.UK PhaseBanner with the message "This is a new service"
        >>> ds.PhaseBanner("This is a new service", phase="Beta")
        # Renders the same PhaseBanner but with a "Beta" tag instead of "Alpha"

    Args:
        *content (FT): The content to display in the phase banner.
        phase (str, optional): The phase of the project. Defaults to "Alpha".
        **kwargs: Additional keyword arguments.

    Returns:
        FT: A FastHTML PhaseBanner component.

    .. _GOV.UK Phase Banner: https://design-system.service.gov.uk/components/phase-banner/
    """
    return fh.Div(
        fh.Div(
            fh.P(
                fh.Strong(
                    phase,
                    cls="govuk-tag govuk-phase-banner__content__tag",
                ),
                fh.Span(
                    *content,
                    cls="govuk-phase-banner__text",
                ),
                cls="govuk-phase-banner__content",
            ),
            cls="govuk-phase-banner",
        ),
        cls="govuk-width-container",
        **kwargs,
    )


def Page(*content, navigation=None, sidebar=None) -> fh.FT:
    """
    A standard `GOV.UK Page`_ component with a GOV.UK Header, a GOV.UK Phase Banner,
    a GOV.UK Footer and - whatever content you might want to add to it.

    The Phase Banner and the Footer are loaded from `/phase` and `/footer` endpoints
    respectively using `htmx-get` directives.

    This is to ensure that there is one and only one definition of a Phase Banner
    and a Footer in your codebase.

    Examples:

        >>> ds.Page("Hello World!")
        # Renders a page with GOV.UK Header and the text "Hello World!"
        # Footer and Phase Banner are loaded dynamically
        >>> ds.Page(ds.H1("Hello World!"))
        # Renders a page with the heading - "Hello World!"
        >>> nav = ds.Navigation("/", "/feedback")
        >>> ds.Page(navigation=nav)
        # Renders a page with the given service navigation component
        >>> ds.Page(sidebar="Hello World!")
        # Renders a page with a sidebar that contains the text - "Hello World!"

    Args:
        *content: List of content for the Page.
        navigation (optional): Navigation component. Defaults to None.
        sidebar (optional): Sidebar content. Defaults to None.

    Returns:
        FT: A FastHTML Page component.

    .. _GOV.UK Page: https://design-system.service.gov.uk/styles/page-template/
    """
    return fh.Body(
        fh.Script(
            "document.body.className += ' js-enabled' + ('noModule' in HTMLScriptElement.prototype ? ' govuk-frontend-supported' : '');",
        ),
        # Cookie banner until the user hides it
        fh.Div(hx_get="/cookie-banner", hx_trigger="load"),
        Header("Fast GOV.UK", "/"),
        # Navigation
        navigation,
        # Phase banner
        fh.Div(hx_get="/phase", hx_trigger="load"),
        fh.Div(
            fh.Main(
                fh.Div(
                    fh.Div(
                        # Every page will check if there are any notifications
                        fh.Div(hx_get="/notifications", hx_trigger="load"),
                        *content,
                        cls="govuk-grid-column-two-thirds",
                    ),
                    fh.Div(*sidebar or [], cls="govuk-grid-column-one-thirds"),
                    cls="govuk-grid-row",
                ),
                cls="govuk-main-wrapper",
            ),
            cls="govuk-width-container",
        ),
        # Get central service footer
        fh.Div(hx_get="/footer", hx_trigger="load"),
        fh.Script(src="/govuk-frontend-5.11.1.min.js", type="module"),
        fh.Script(
            "import {initAll} from '/govuk-frontend-5.11.1.min.js'; initAll();",
            type="module",
        ),
    )


def Cookies(*content: fh.FT):
    """
    A standard `GOV.UK Cookie Page`_ component that includes definition of cookies and
    a table detailing the 2 essential cookies used by fast-gov-uk services.

    Examples:

        >>> ds.Cookies()
        # Render the Cookies page
        >>> ds.Cookies(ds.H2("Additional Cookies"), ...)
        # Renders the Cookies page with additional cookies

    This is just the component. You would probably want to render the Cookies
    page at the `/cookies` URL. You can do so by adding the following function
    to your `app.py`:

    .. code-block:: python

        @fast.page
        def cookies():
            return ds.Cookies()

    Args:
        *content (FT): List of content for the Page.

    Returns:
        FT: A FastHTML Cookies page component.

    .. _GOV.UK Cookie Page: https://design-system.service.gov.uk/patterns/cookies-page/
    """
    return Page(
        H1("Cookies"),
        P("Cookies are small files saved on your phone, tablet or computer when you visit a website."),
        P("We use cookies to make this site work and collect information about how you use our service."),
        H2("Essential Cookies"),
        P("Essential cookies keep your information secure while you use this service. We do not need to ask permission to use them."),
        Table(
            data=[
                {"name": "session_cookie", "purpose": "Used to store your settings and progress", "expires": "1 day"},
                {"name": "cookie_policy", "purpose": "Saves your cookie consent settings", "expires": "1 year"},
            ],
            caption="Essential cookies we use",
        ),
        *content,
    )



import pytest

from fast_gov_uk.design_system import H1, H2, H3, A, Li, P, Ul


@pytest.mark.parametrize(
    "kwargs, expected",
    (
        (
            {"text": "Test"},
            '<a href="#" class="govuk-link">Test</a>',
        ),
        (
            {"text": "Test", "href": "/test"},
            '<a href="/test" class="govuk-link">Test</a>',
        ),
        (
            {"text": "Test", "inverse": True},
            '<a href="#" class="govuk-link govuk-link--inverse">Test</a>',
        ),
        (
            {"text": "Test", "visited": True},
            '<a href="#" class="govuk-link govuk-link--visited">Test</a>',
        ),
        (
            {"text": "Test", "newtab": True},
            '<a href="#" target="_blank" rel="noopener noreferrer" class="govuk-link">Test</a>',
        ),
    ),
)
def test_a(kwargs, expected):
    """Test A with various parameters.
    Args:
        kwargs (dict): The arguments to pass to A.
        expected (str): The expected HTML output.
    """
    link = A(**kwargs)
    assert str(link) == expected


@pytest.mark.parametrize(
    "kwargs, expected",
    (
        (
            {"text": "Test"},
            '<h1 class="govuk-heading-l">Test</h1>',
        ),
        (
            {"text": "Test", "size": "m"},
            '<h1 class="govuk-heading-m">Test</h1>',
        ),
        (
            {"text": "Test", "caption": "Caption"},
            '<h1 class="govuk-heading-l">Test<span class="govuk-caption-l">Caption</span></h1>',
        ),
    ),
)
def test_h1(kwargs, expected):
    """Test H1 with various parameters.
    Args:
        kwargs (dict): The arguments to pass to H1.
        expected (str): The expected HTML output.
    """
    heading = H1(**kwargs)
    assert str(heading) == expected


@pytest.mark.parametrize(
    "kwargs, expected",
    (
        (
            {"text": "Test"},
            '<h2 class="govuk-heading-m">Test</h2>',
        ),
        (
            {"text": "Test", "size": "s"},
            '<h2 class="govuk-heading-s">Test</h2>',
        ),
        (
            {"text": "Test", "caption": "Caption"},
            '<h2 class="govuk-heading-m">Test<span class="govuk-caption-m">Caption</span></h2>',
        ),
    ),
)
def test_h2(kwargs, expected):
    """Test H2 with various parameters.
    Args:
        kwargs (dict): The arguments to pass to H2.
        expected (str): The expected HTML output.
    """
    heading = H2(**kwargs)
    assert str(heading) == expected


@pytest.mark.parametrize(
    "kwargs, expected",
    (
        (
            {"text": "Test"},
            '<h3 class="govuk-heading-s">Test</h3>',
        ),
        (
            {"text": "Test", "size": "m"},
            '<h3 class="govuk-heading-m">Test</h3>',
        ),
        (
            {"text": "Test", "caption": "Caption"},
            '<h3 class="govuk-heading-s">Test<span class="govuk-caption-s">Caption</span></h3>',
        ),
    ),
)
def test_h3(kwargs, expected):
    """Test H3 with various parameters.
    Args:
        kwargs (dict): The arguments to pass to H3.
        expected (str): The expected HTML output.
    """
    heading = H3(**kwargs)
    assert str(heading) == expected


@pytest.mark.parametrize(
    "kwargs, expected",
    (
        (
            {"content": "Test"},
            '<p class="govuk-body">Test</p>',
        ),
        (
            {"content": "Test", "lead": True},
            '<p class="govuk-body-l">Test</p>',
        ),
        (
            {"content": "Test", "small": True},
            '<p class="govuk-body-s">Test</p>',
        ),
        (
            {"content": ["Test ", A("a link", "/")]},
            '<p class="govuk-body">Test <a href="/" class="govuk-link">a link</a></p>',
        ),
    ),
)
def test_p(kwargs, expected):
    """Test P with various parameters.
    Args:
        kwargs (dict): The arguments to pass to P.
        expected (str): The expected HTML output.
    """
    para = P(*kwargs.pop("content"), **kwargs)
    assert str(para) == expected


@pytest.mark.parametrize(
    "args, kwargs, expected",
    (
        (
            (Li("Item 1"), Li("Item 2")),
            {},
            '<ul class="govuk-list"><li>Item 1</li><li>Item 2</li></ul>',
        ),
        (
            (Li("Item 1"), Li("Item 2")),
            {"bullet": True},
            '<ul class="govuk-list govuk-list--bullet"><li>Item 1</li><li>Item 2</li></ul>',
        ),
        (
            (Li("Item 1"), Li("Item 2")),
            {"numbered": True},
            '<ul class="govuk-list govuk-list--number"><li>Item 1</li><li>Item 2</li></ul>',
        ),
        (
            (Li("Item 1"), Li("Item 2")),
            {"spaced": True},
            '<ul class="govuk-list govuk-list--spaced"><li>Item 1</li><li>Item 2</li></ul>',
        ),
    ),
)
def test_ul(args, kwargs, expected):
    """Test UL with various parameters.
    Args:
        args (list): The Lis to pass to Ul.
        kwargs (dict): The arguments to pass to Ul.
        expected (str): The expected HTML output.
    """
    list = Ul(*args, **kwargs)
    assert str(list) == expected


@pytest.mark.parametrize("component", (
    A("test", hx_test="foo"),
    H1("test", hx_test="foo"),
    H2("test", hx_test="foo"),
    H3("test", hx_test="foo"),
    P("test", hx_test="foo"),
    Ul("test", hx_test="foo"),
))
def test_html_attribute(component, html):
    """
    Test that passes an html attribute to components - that is not
    explicitly handled but should be passed through to the
    underlying FT.
    """
    assert 'hx-test="foo"' in html(component)


def test_p_error():
    with pytest.raises(ValueError, match="Cannot set both lead and small to True."):
        P("test", lead=True, small=True)


def test_ul_error():
    with pytest.raises(ValueError, match="Cannot set both bullet and numbered to True."):
        Ul("test", bullet=True, numbered=True)



from os import environ as env

from fast_gov_uk import Fast, serve
import fast_gov_uk.design_system as ds

# This is your settings. They are set for development on your
# own computer by default. When you are deploying your service,
# you can override them by setting environment variables.
SETTINGS = {
    "SERVICE_NAME": env.get("SERVICE_NAME", "Fast Gov UK"),
    "DATABASE_URL": env.get("DATABASE_URL", "data/service.db"),
    "DEV_MODE": env.get("DEV_MODE", True),
    "NOTIFY_API_KEY": env.get("NOTIFY_API_KEY", None),
}

# This creates the Fast object that encapsulates everything
# in your service
fast = Fast(SETTINGS)


# If I do @fast.page() instead, this page will be available
# on /home i.e. the name of the function
@fast.page("/")
def home():
    return ds.Page(
        # A single Paragraph
        ds.P("Welcome to Fast Gov UK.")
    )


# Serves the app
serve(app="fast")



"""
GOV.UK Design System form input components including TextInput, Textarea,
Select, Checkboxes, Radios etc.

These components are subclasses of :py:class:`Field` which provides common
functionality like rendering labels, hints, errors, setting values, and
validating required fields.
"""

from datetime import date
from enum import Enum
from pathlib import Path
from typing import List

import fasthtml.common as fh

from .typography import H2, A
from .utils import mkid


def Label(
    field_id: str,
    text: str,
    heading: str = "",
    required: bool = True,
    extra_cls: str = "",
) -> fh.FT:
    """
    Used by the :py:class:`Field` component to render a label.

    Handles whether the label is a heading, whether the field is required etc.

    Use this if you are creating a custom field.

    Args:
        field_id (str): HTML id of the field this label is for.
        text (str): Text to be displayed in the label.
        heading (str, optional): Heading size. Defaults to "".
        required (bool, optional): Is this for a field that is required? Defaults to True.
        extra_cls (str, optional): Extra CSS classes. Defaults to "".

    Returns:
        FT: A FastHTML label component.
    """
    optional = "" if required else " (Optional)"
    heading_cls = f" govuk-label--{heading}" if heading else ""
    label = fh.Label(
        f"{text}{optional}", cls=f"govuk-label{heading_cls}{extra_cls}", _for=field_id
    )
    if heading:
        label = fh.H1(label, cls="govuk-label-wrapper")
    return label


def Hint(field_id: str, text: str, extra_cls: str = "") -> fh.FT:
    """
    Used by the :py:class:`Field` component to render a hint.

    Use this if you are creating a custom field.

    Args:
        field_id (str): HTML id of the field this hint is for.
        text (str): Text to be displayed as hint.
        extra_cls (str, optional): Extra CSS classes. Defaults to "".

    Returns:
        FT: A FastHTML hint component.
    """
    return fh.Div(text, cls=f"govuk-hint{extra_cls}", id=f"{field_id}-hint")


def Error(field_id: str, text: str, extra_cls: str = "") -> fh.FT:
    """
    Used by the :py:class:`Field` component to render error message.

    Use this if you are creating a custom field.

    Args:
        field_id (str): HTML id of the field this error is for.
        text (str): Text to be displayed as error.
        extra_cls (str, optional): Extra CSS classes. Defaults to "".

    Returns:
        FT: A FastHTML error component.
    """
    return fh.P(
        fh.Span("Error: ", cls="govuk-visually-hidden"),
        text,
        cls=f"govuk-error-message{extra_cls}",
        id=f"{field_id}-error",
    )


class AbstractField:
    """
    Every form field is a subclass of this type.

    This is useful when e.g. we are having to distinguish fields in a
    form that contains fields as well as non-field elements like H1 and P.

    These types of mixed forms are quite common in question pages.
    """
    pass


class Field(AbstractField):
    """
    Baseclass for form inputs. Provides scaffolding for -

        - rendering labels
        - rendering hints
        - rendering errors
        - setting field values
        - validation for required fields

    If you are want to implement a custom GOV.UK input, this is a good
    base class to inherit from.

    Args:
        name (str): The name of the field.
        label (str, optional): Label for the field. Defaults to "".
        hint (str, optional): Hint for the field. Defaults to "".
        error (str, optional): Error message for the field. Defaults to "".
        heading (str, optional): Heading size. Defaults to "".
        required (bool, optional): Is this field required? Defaults to True.
        **kwargs: Additional keyword arguments.
    """

    def __init__(
        self,
        name: str,
        label: str = "",
        hint: str = "",
        error: str = "",
        heading: str = "",
        required: bool = True,
        **kwargs,
    ):
        self.name = name
        self.label = label
        self.hint = hint
        self.error = error
        self.heading = heading
        self.required = required
        self._value = None
        self.kwargs = kwargs

    @property
    def value(self):
        return self._value

    @property
    async def clean(self):
        """
        Get "clean" data from a field when a form is being processed
        by a form Backend.
        """
        if not self.value:
            return None
        return self.value

    @value.setter
    def value(self, value):
        """
        Assigns a value to a field, runs validation, and sets errors.

        Args:
            value: The value to assign.
        """
        if self.required and not value:
            self.error = "This field is required."
        self._value = value

    @property
    def _id(self):
        """
        Computes a good id from the field name.

        Returns:
            str: Computed id.
        """
        return mkid(self.name)

    @property
    def annotations(self):
        """
        Returns label, hint, and error components for the field.

        Returns:
            tuple: (Label, Hint, Error)
        """
        return (
            (
                Label(self._id, self.label, self.heading, self.required)
                if self.label
                else ""
            ),
            Hint(self._id, self.hint) if self.hint else "",
            Error(self._id, self.error) if self.error else "",
        )

    def __ft__(self, *children: fh.FT, **kwargs) -> fh.FT:
        """
        Renders the field as a FastHTML component.

        Args:
            *children (FT): Optional children components to include.
            **kwargs: Additional keyword arguments.

        Returns:
            FT: A FastHTML component for the field.
        """
        extra_cls = kwargs.pop("cls", "")
        error_cls = " govuk-form-group--error" if self.error else ""
        return fh.Div(
            *self.annotations,
            *children,
            cls=f"govuk-form-group{extra_cls}{error_cls}",
            **kwargs,
        )


class Select(Field):
    """
    `GOV.UK Select`_ component. Renders a dropdown. Inherits from Field.

    This component should only be used as a last resort in public-facing services because research
    shows that some users find selects very difficult to use.

    Use Checkbox or Radio components instead.

    Examples:

        >>> ds.Select("question", choices={"yes": "Yes", "no": "No"})
        # Renders a dropdown select with the given options - "Yes" and "No.
        # Note the values for these options are "yes" and "no" respectively

    Args:
        *args: Arguments for Field.
        choices (dict, optional): Choices for the select. Defaults to None.
        **kwargs: Additional keyword arguments.

    .. _GOV.UK Select: https://design-system.service.gov.uk/components/select/
    """

    def __init__(self, *args, choices: dict | None = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.choices = choices or {}

    @property
    async def clean(self):
        """
        Get "clean" data from a field when a form is being processed
        by a form Backend.
        """
        for val, text in self.choices.items():
            if val == self.value:
                return text

    @property
    def options(self):
        """
        Convert the choices dict into a list of <option> to be
        inserted into the Select component.
        """
        return [
            fh.Option(text, value=value, selected=(value == self.value))
            for value, text in self.choices.items()
        ]

    def __ft__(self, *children, **kwargs) -> fh.FT:
        """
        Renders the select field as a FastHTML component.

        Args:
            *children: Optional children components.
            **kwargs: Additional keyword arguments.

        Returns:
            FT: FastHTML Select component.
        """
        error_cls = " govuk-select--error" if self.error else ""
        return super().__ft__(
            fh.Select(
                name=self.name,
                *self.options,
                _id=self._id,
                cls=f"govuk-select{error_cls}",
            ),
            **self.kwargs,
        )


class Textarea(Field):
    """
    `GOV.UK Textarea`_ component. Inherits from Field.

    Use this component when you need to let users enter more than a single line of text.

    Examples:

        >>> ds.Textarea('test')
        # Renders a text area with the default 5 rows

    Args:
        *args: Arguments for Field.
        rows (int, optional): Number of rows in the textarea. Defaults to 5.
        **kwargs: Additional keyword arguments.

    .. _GOV.UK Textarea: https://design-system.service.gov.uk/components/textarea/
    """

    def __init__(self, *args, rows: int = 5, **kwargs):
        super().__init__(*args, **kwargs)
        self.rows = rows

    def __ft__(self, *children, **kwargs) -> fh.FT:
        """
        Renders the textarea field as a FastHTML component.

        Args:
            *children: Optional children components.
            **kwargs: Additional keyword arguments.

        Returns:
            FT: FastHTML Textarea component.
        """
        error_cls = " govuk-textarea--error" if self.error else ""
        return super().__ft__(
            fh.Textarea(
                name=self.name,
                rows=self.rows,
                value=self.value,
                id=self._id,
                aria_describedby=f"{self._id}-hint {self._id}-error",
                cls=f"govuk-textarea{error_cls}",
            ),
            **self.kwargs,
        )


class PasswordInput(Field):
    """
    `GOV.UK Password input`_ component. Inherits from Field.

    Use this component when you need users to type a password.

    Examples:

        >>> ds.PasswordInput('test')
        # Renders a password input

    Args:
        *args: Arguments for Field.
        **kwargs: Additional keyword arguments.

    .. _GOV.UK Password input: https://design-system.service.gov.uk/components/password-input/
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def input(self):
        """
        Input part for this component.
        """
        return fh.Input(
            type="password",
            name=self.name,
            id=self._id,
            aria_describedby=f"{self._id}-hint {self._id}-error",
            cls=(
                "govuk-input govuk-password-input__input "
                "govuk-js-password-input-input"
                f"{' govuk-input--error' if self.error else ''}"
            ),
        )

    @property
    def button(self):
        """
        Button part for this component
        """
        return fh.Button(
            "Show",
            type="button",
            cls=(
                "govuk-button govuk-button--secondary "
                "govuk-password-input__toggle "
                "govuk-js-password-input-toggle"
            ),
            data_module="govuk-password-input__toggle",
            aria_controls=self._id,
            aria_label="Show password",
            hidden=True,
        )

    def __ft__(self, *children, **kwargs) -> fh.FT:
        """
        Renders the password input field as a FastHTML component.

        Args:
            *children: Optional children components.
            **kwargs: Additional keyword arguments.

        Returns:
            FT: FastHTML component for PasswordInput.
        """
        return super().__ft__(
            fh.Div(
                self.input,
                self.button,
                cls="govuk-input__wrapper govuk-password-input__wrapper",
            ),
            cls=" govuk-password-input",
            data_module="govuk-password-input",
            **self.kwargs,
        )


class CharacterCount(Field):
    """
    `GOV.UK Character count`_ component. Renders a Textarea with a character/word count message.

    Help users know how much text they can enter when there is a limit on the number of characters.

    It is recommended to always test your service without a character count first and to only use the
    character count component when there is a good reason for limiting the number of characters users
    can enter.

    Examples:

        >>> ds.CharacterCount('test')
        # Renders a text area
        >>> ds.CharacterCount('test', maxchars=100)
        # Renders a text area with a limit of 100 characters max
        >>> ds.CharacterCount('test', maxwords=10)
        # Renders a text area with a limit of 10 words max
        >>> ds.CharacterCount('test', maxwords=10, threshold=50)
        # Renders a text area with a limit of 10 words max but the message that says
        # "You have X words left" only appears after 50% of the limit is crossed.

    Args:
        *args: Arguments for Field.
        rows (int, optional): Number of rows in the textarea. Defaults to 5.
        maxchars (int, optional): Max characters allowed. Defaults to None.
        maxwords (int, optional): Max words allowed. Defaults to None.
        threshold (int, optional): Threshold percent for showing count message. Defaults to None.
        **kwargs: Additional keyword arguments.

    .. _GOV.UK Character count: https://design-system.service.gov.uk/components/character-count/
    """

    def __init__(
        self,
        *args,
        rows: int = 5,
        maxchars: int | None = None,
        maxwords: int | None = None,
        threshold: int | None = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.rows = rows
        self.maxchars = maxchars
        self.maxwords = maxwords
        self.threshold = threshold

    @Field.value.setter
    def value(self, value):
        """
        Assign error if user have exceeded maxchars or maxwords.
        """
        self._value = value
        if self.required and not value:
            self.error = "This field is required."
            return
        if self.maxchars:
            if len(self._value) > self.maxchars:
                self.error = f"Characters exceed limit of {self.maxchars}."
        if self.maxwords:
            words = self._value.split()
            if len(words) > self.maxwords:
                self.error = f"Words exceed limit of {self.maxwords}."

    @property
    def textarea(self):
        """
        Textarea part of CharacterCount.
        """
        error_cls = " govuk-textarea--error" if self.error else ""
        return fh.Textarea(
            name=self.name,
            rows=self.rows,
            value=self.value,
            id=self._id,
            aria_describedby=f"{self._id}-hint {self._id}-error",
            cls=f"govuk-textarea govuk-js-character-count{error_cls}",
        )

    @property
    def message(self):
        """
        Message part of CharacterCount.
        """
        message = (
            f"You can enter up to {self.maxchars} characters."
            if self.maxchars
            else f"You can enter up to {self.maxwords} words."
        )
        return fh.Div(
            message,
            cls="govuk-hint govuk-character-count__message",
            id=f"{self._id}-info",
        )

    def __ft__(self, *children, **kwargs) -> fh.FT:
        """
        Renders the character count field as a FastHTML component.

        Args:
            *children: Optional children components.
            **kwargs: Additional keyword arguments.

        Returns:
            FT: FastHTML component for CharacterCount.
        """
        return super().__ft__(
            self.textarea,
            self.message,
            cls=" govuk-character-count",
            data_module="govuk-character-count",
            data_maxlength=self.maxchars,
            data_maxwords=self.maxwords,
            data_threshold=self.threshold,
            **self.kwargs,
        )


class InputWidth(Enum):
    """
    Enum to define GOV.UK supported input widths to be passed into
    :py:class:`TextInput` as `width` param.
    """
    DEFAULT = 0
    W20 = 1
    W10 = 2
    W5 = 3
    W4 = 4
    W3 = 5
    W2 = 6
    WFULL = 7
    W75 = 8
    W66 = 9
    W50 = 10
    W33 = 11
    W25 = 12


class TextInput(Field):
    """
    `GOV.UK Text Input`_ component. Inherits from Field.

    Use the text input component when you need to let users enter text that is no longer than a
    single line, such as their name or email address.

    Examples:

        >>> ds.TextInput('test')
        # Renders a standard text input
        >>> ds.TextInput('test', width=InputWidth.W50)
        # Renders a text input with a width of 50%
        >>> ds.TextInput('test', prefix="£")
        # Renders a text input with a prefix of "£"
        >>> ds.TextInput('test', suffix="%")
        # Renders a text input with a suffix of "%"
        >>> ds.TextInput('test', autocomplete="postal-code")
        # Renders a text input that uses the autocomplete features in
        # modern browsers to fill-in the postcode
        >>> ds.TextInput('test', numeric=True)
        # Renders a text input that uses the numeric keypad on devices
        # with on-screen keyboards.
        >>> ds.TextInput('test', spellcheck=True)
        # Renders a text input with the spellcheck turned on
        >>> ds.TextInput('test', extraspacing=True)
        # Renders a text input with extra spacing

    Args:
        *args: Arguments for Field.
        width (InputWidth, optional): Width of TextInput. Defaults to InputWidth.DEFAULT.
        prefix (str, optional): Prefix to TextInput. Defaults to "".
        suffix (str, optional): Suffix to TextInput. Defaults to "".
        autocomplete (str, optional): Autocomplete value. Defaults to "".
        numeric (bool, optional): Is TextInput numeric? Defaults to False.
        spellcheck (bool, optional): Enable spellcheck. Defaults to False.
        extra_spacing (bool, optional): Extra letter spacing. Defaults to False.
        **kwargs: Additional keyword arguments.

    .. _GOV.UK Text Input: https://design-system.service.gov.uk/components/text-input/
    """

    def __init__(
        self,
        *args,
        width: InputWidth = InputWidth.DEFAULT,
        prefix: str = "",
        suffix: str = "",
        autocomplete: str = "",
        numeric: bool = False,
        spellcheck: bool = False,
        extra_spacing: bool = False,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.width = width
        self.prefix = prefix
        self.suffix = suffix
        self.autocomplete = autocomplete
        self.numeric = numeric
        self.spellcheck = spellcheck
        self.extra_spacing = extra_spacing
        self.width_cls = {
            InputWidth.DEFAULT: "",
            InputWidth.W20: " govuk-input--width-20",
            InputWidth.W10: " govuk-input--width-10",
            InputWidth.W5: " govuk-input--width-5",
            InputWidth.W4: " govuk-input--width-4",
            InputWidth.W3: " govuk-input--width-3",
            InputWidth.W2: " govuk-input--width-2",
            InputWidth.WFULL: " govuk-!-width-full",
            InputWidth.W75: " govuk-!-width-three-quarters",
            InputWidth.W66: " govuk-!-width-two-thirds",
            InputWidth.W50: " govuk-!-width-one-half",
            InputWidth.W33: " govuk-!-width-one-third",
            InputWidth.W25: " govuk-!-width-one-quarter",
        }

    @property
    def input(self):
        """
        Input part of this component.
        """
        error_cls = " govuk-input--error" if self.error else ""
        spacing_cls = " govuk-input--extra-letter-spacing" if self.extra_spacing else ""
        return fh.Input(
            type="text",
            name=self.name,
            value=self.value,
            id=self._id,
            aria_describedby=f"{self._id}-hint {self._id}-error",
            cls=f"govuk-input{self.width_cls[self.width]}{error_cls}{spacing_cls}",
            inputmode="numeric" if self.numeric else None,
            spellcheck=self.spellcheck,
            autocomplete=self.autocomplete,
        )

    def __ft__(self, *children, **kwargs) -> fh.FT:
        """
        Renders the text input field as a FastHTML component, including prefix/suffix.

        Args:
            *children: Optional children components.
            **kwargs: Additional keyword arguments.

        Returns:
            FT: FastHTML component for TextInput.
        """
        input = self.input
        if self.prefix:
            input = fh.Div(
                fh.Div(self.prefix, cls="govuk-input__prefix", aria_hidden=True),
                self.input,
                cls="govuk-input__wrapper",
            )
        if self.suffix:
            input = fh.Div(
                self.input,
                fh.Div(self.suffix, cls="govuk-input__suffix", aria_hidden=True),
                cls="govuk-input__wrapper",
            )
        return super().__ft__(input, **self.kwargs)


class Checkbox(AbstractField):
    """
    Checkbox component. Used to define individual checkboxes to be passed into
    the :py:class:`Checkboxes` component.

    Note that this inherits from :py:class:`AbstractField` instead of :py:class:`Field` because
    it is only meant to be used inside the :py:class:`Checkboxes` component - which inherits from
    :py:class:`Field`.

    Examples:

        >>> cb = ds.Checkbox('test', 'test', 'Test Label')
        # Renders a Checkbox called "test" with a label "Test Label"
        >>> cb = ds.Checkbox('test', 'test', 'Test Label', checked=True)
        # Renders the same Checkbox but its checked by default
        >>> cb = ds.Checkbox('test', 'test', 'Test Label', checked=True)
        # Renders the same Checkbox but its checked by default
        >>> cb = ds.Checkbox('test', 'test', 'Test Label', exclusive=True)
        # Renders the same Checkbox when this checkbox it clicked, it unchecks all other
        # checkboxes - handy for options like "None of the above"

    Args:
        name (str): The name of the checkbox element.
        value (str): The value of the checkbox element.
        label (str): Label for the checkbox element.
        hint (str, optional): Hint for the checkbox element. Defaults to "".
        checked (bool, optional): Make this checkbox checked. Defaults to False.
        exclusive (bool, optional): Make this checkbox exclusive. Defaults to False.
        **kwargs: Additional keyword arguments.
    """

    def __init__(
        self,
        name: str,
        value: str,
        label: str,
        hint: str = "",
        checked: bool = False,
        exclusive: bool = False,
        **kwargs,
    ):
        super().__init__()
        self.name = name
        self.value = value
        self.label = label
        self.hint = hint
        self.checked = checked
        self.exclusive = exclusive
        self.kwargs = kwargs

    @property
    def _id(self):
        """
        Checkbox id from name and value.
        """
        return f"{mkid(self.name)}-{self.value}"

    @property
    def label_component(self):
        """
        Label component.
        """
        return Label(self._id, self.label, extra_cls=" govuk-checkboxes__label")

    @property
    def hint_component(self):
        """
        Hint component.
        """
        return Hint(self._id, self.hint, extra_cls=" govuk-checkboxes__hint")

    def __ft__(self, *children, **kwargs) -> fh.FT:
        """
        Renders the checkbox as a FastHTML component.

        Args:
            *children: Optional children components.
            **kwargs: Additional keyword arguments.

        Returns:
            FT: FastHTML Checkbox component.
        """
        return fh.Div(
            fh.Input(
                self.label_component,
                self.hint_component if self.hint else "",
                _id=self._id,
                name=self.name,
                type="checkbox",
                value=self.value,
                checked=self.checked,
                cls="govuk-checkboxes__input",
                data_behaviour="exclusive" if self.exclusive else None,
            ),
            cls="govuk-checkboxes__item",
            **self.kwargs,
        )


class Checkboxes(Field):
    """
    `GOV.UK Checkboxes`_ component. Let users select one or more options by using the checkboxes
    component.

    Use the checkboxes component when you need to help users: select multiple options from a list
    or to toggle a single option on or off.

    Use the handy "choices" parameter as a shortcut to quickly define standard checkbox groups.

    Examples:

        >>> ds.Checkboxes("question", choices={"yes": "Yes!", "no": "Maybe later.."})
        # Renders checkboxes called "question" with values "yes" and "no" with texts
        # "Yes!" and "Maybe later.." respectively
        >>> ds.Checkboxes("question", choices={"yes": "Yes!", "no": "Maybe later.."}, small=True)
        # Renders the same checkbox with more compact styling
        >>> cb_yes = ds.Checkbox("question", "yes", "Yes!")
        >>> cb_no = ds.Checkbox("question", "no", "Maybe later..")
        >>> ds.Checkboxes("question", cb_yes, cb_no)
        # Renders the same checkboxes but by defining individual components, we can use
        # all of the underlying configurations

    Args:
        *args: Arguments for Field.
        checkboxes (List[Checkbox], optional): List of Checkbox components.
        choices (dict, optional): Shorthand for simple checkboxes.
        small (bool, optional): Renders small Checkboxes. Defaults to False.
        **kwargs: Additional keyword arguments.

    .. _GOV.UK Checkboxes: https://design-system.service.gov.uk/components/checkboxes/
    """

    def __init__(
        self,
        *args,
        checkboxes: List[Checkbox] | None = None,
        choices: dict | None = None,
        small: bool = False,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.checkboxes = checkboxes or []
        self.choices = choices or {}
        self.small = small
        if self.choices:
            self.make_checkboxes()

    @property
    async def clean(self):
        """
        Get "clean" data from a field when a form is being processed
        by a form Backend.
        """
        for cb in self.checkboxes:
            if cb.value == self.value:
                return cb.label

    def make_checkboxes(self):
        """
        Make checkboxes from choices.
        """
        for value, label in self.choices.items():
            cb = Checkbox(self.name, value, label)
            self.checkboxes.append(cb)

    def __ft__(self, *children, **kwargs) -> fh.FT:
        """
        Renders the checkboxes field as a FastHTML component.

        Args:
            *children: Optional children components.
            **kwargs: Additional keyword arguments.

        Returns:
            FT: FastHTML Checkboxes component.
        """
        small_cls = " govuk-checkboxes--small" if self.small else ""
        for check in self.checkboxes:
            check.checked = check.value == self.value
        return super().__ft__(
            fh.Fieldset(
                fh.Div(
                    *self.checkboxes,
                    cls=f"govuk-checkboxes{small_cls}",
                    data_module="govuk-checkboxes",
                ),
                cls="govuk-fieldset",
                aria_describedby=f"{self._id}-hint",
                id=self._id,
            ),
            **self.kwargs,
        )


class Radio(AbstractField):
    """
    Radio component. Used to define individual radioes to be passed into
    the :py:class:`Radios` component.

    Note that this inherits from :py:class:`AbstractField` instead of :py:class:`Field` because
    it is only meant to be used inside the :py:class:`Radios` component - which inherits from
    :py:class:`Field`.

    Examples:

        >>> cb = ds.Radio('test', 'test', 'Test Label')
        # Renders a radio called "test" with a label "Test Label"
        >>> cb = ds.Radio('test', 'test', 'Test Label', checked=True)
        # Renders the same radio but its checked by default
        >>> cb = ds.Radio('test', 'test', 'Test Label', checked=True)
        # Renders the same radio but its checked by default

    Args:
        name (str): The name of the radio element.
        value (str): The value of the radio element.
        label (str): Label for the radio element.
        hint (str, optional): Hint for the radio element. Defaults to "".
        checked (bool, optional): Make this radio checked. Defaults to False.
        reveal (Field, optional): Field revealed when Radio is selected. Defaults to None.
        **kwargs: Additional keyword arguments.
    """

    def __init__(
        self,
        name: str,
        value: str,
        label: str,
        hint: str = "",
        checked: bool = False,
        reveal: Field | None = None,
        **kwargs,
    ):
        super().__init__()
        self.name = name
        self.value = value
        self.label = label
        self.hint = hint
        self.checked = checked
        self.reveal = reveal
        self.kwargs = kwargs

    @property
    def _id(self):
        """
        Computes radio id from name and value.
        """
        return f"{mkid(self.name)}-{self.value}"

    @property
    def label_component(self):
        """
        Label component.
        """
        return Label(self._id, self.label, extra_cls=" govuk-radios__label")

    @property
    def hint_component(self):
        """
        Hint component.
        """
        return Hint(self._id, self.hint, extra_cls=" govuk-radios__hint")

    @property
    def data_aria_controls(self):
        """
        Aria-controls component.
        """
        if self.reveal:
            return f"conditional-{self._id}"
        return None

    @property
    def base_radio(self):
        """
        Base radio component.
        """
        return fh.Div(
            fh.Input(
                self.label_component,
                self.hint_component if self.hint else "",
                _id=self._id,
                name=self.name,
                type="radio",
                value=self.value,
                checked=self.checked,
                cls="govuk-radios__input",
                data_aria_controls=self.data_aria_controls,
            ),
            cls="govuk-radios__item",
        )

    @property
    def _reveal(self):
        """
        Reveal component.
        """
        return fh.Div(
            self.reveal,
            cls="govuk-radios__conditional govuk-radios__conditional--hidden",
            id=f"conditional-{self._id}",
        )

    def __ft__(self, *children, **kwargs) -> fh.FT:
        """
        Renders the radio as a FastHTML component.

        Args:
            *children: Optional children components.
            **kwargs: Additional keyword arguments.

        Returns:
            FT: FastHTML Radio component.
        """
        return (
            fh.Div(
                self.base_radio,
                self._reveal,
                **self.kwargs,
            )
            if self.reveal
            else self.base_radio
        )


class Radios(Field):
    """
    `GOV.UK Radios`_ component. Use the radios component when users can only select one option
    from a list.

    Use the handy "choices" parameter as a shortcut to quickly define standard radio groups.

    Examples:

        >>> ds.Radios("question", choices={"yes": "Yes!", "no": "Maybe later.."})
        # Renders radios called "question" with values "yes" and "no" with texts
        # "Yes!" and "Maybe later.." respectively
        >>> ds.Radios("question", choices={"yes": "Yes!", "no": "Maybe later.."}, small=True)
        # Renders the same checkbox with more compact styling
        >>> ds.Radios("question", choices={"yes": "Yes!", "no": "Maybe later.."}, inline=True)
        # Renders the same checkbox with inline styling
        >>> yes = ds.Radio("question", "yes", "Yes!")
        >>> no = ds.Radio("question", "no", "Maybe later..")
        >>> ds.Radios("question", yes, no)
        # Renders the same radios but by defining individual components, we can use
        # all of the underlying configurations

    Args:
        *args: Arguments for Field.
        radios (List[Radio], optional): List of Radio components.
        choices (dict, optional): Shorthand for simple radios.
        small (bool, optional): Renders small Radios. Defaults to False.
        inline (bool, optional): Renders inline Radios. Defaults to False.
        **kwargs: Additional keyword arguments.

    .. _GOV.UK Radios: https://design-system.service.gov.uk/components/radios/
    """

    def __init__(
        self,
        *args,
        radios: List[Radio] | None = None,
        choices: dict | None = None,
        small: bool = False,
        inline: bool = False,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.radios = radios or []
        self.choices = choices or {}
        self.small = small
        self.inline = inline
        if self.choices:
            self.make_radios()

    @property
    async def clean(self):
        """
        Get "clean" data from a field when a form is being processed
        by a form Backend.
        """
        for radio in self.radios:
            if radio.value == self.value:
                return radio.label

    def make_radios(self):
        """
        Make radios from choices.
        """
        for value, label in self.choices.items():
            radio = Radio(self.name, value, label)
            self.radios.append(radio)

    def insert_divider(self):
        """
        Insert "or" if there are more than 2 radios.
        """
        if len(self.radios) <= 2:
            return self.radios
        divider = fh.Div("or", cls="govuk-radios__divider")
        self.radios.insert(-1, divider)
        return self.radios

    def __ft__(self, *children, **kwargs) -> fh.FT:
        """
        Renders the radios field as a FastHTML component.

        Args:
            *children: Optional children components.
            **kwargs: Additional keyword arguments.

        Returns:
            FT: FastHTML Radios component.
        """
        radios = self.insert_divider() or []
        small_cls = " govuk-radios--small" if self.small else ""
        inline_cls = " govuk-radios--inline" if self.inline else ""
        for radio in self.radios:
            radio.checked = radio.value == self.value
        return super().__ft__(
            fh.Fieldset(
                fh.Div(
                    *radios,
                    cls=f"govuk-radios{small_cls}{inline_cls}",
                    data_module="govuk-radios",
                ),
                cls="govuk-fieldset",
                aria_describedby=f"{self._id}-hint",
                id=self._id,
            ),
            **self.kwargs,
        )


class FileUpload(Field):
    """
    `GOV.UK File upload`_ component. Renders a file upload field to help users select and
    upload a file.

    To upload a file, the user can either: use the "Choose file" button or drag and drop a
    file into the file upload input area.

    It is recommended that you should only ask users to upload something if it's critical to
    the delivery of your service.

    Examples:

        >>> ds.FileUpload("test")
        # Renders a file upload field called "test"

    Args:
        *args: Arguments for Field.
        **kwargs: Additional keyword arguments.

    .. _GOV.UK File upload: https://design-system.service.gov.uk/components/radios/
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def valid_file(self):
        """
        Is the uploaded file valid?
        """
        filename = getattr(self.value, "filename", None)
        return filename is not None

    @Field.value.setter
    def value(self, value):
        """
        Checks if the file is valid and assign error accordingly.
        """
        self._value = value
        if self.required and not self.valid_file:
            self.error = "This field is required."
            return

    @property
    async def clean(self):
        """
        Get "clean" data from a field when a form is being processed
        by a form Backend.
        """
        # field was not required
        if not self.value:
            return None
        # Sometimes, when field is left empty
        # we get a reference to an empty UploadFile
        # TODO: Figure out when and why this happens
        # and write tests against it.
        if not self.valid_file:
            return None
        try:
            buffer = await self.value.read()
            filename = self.value.filename
            path = Path("media") / filename
            path.write_bytes(buffer)
            return str(path)
        except (ValueError, AttributeError, OSError):
            raise

    def __ft__(self, *children, **kwargs) -> fh.FT:
        """
        Renders the file upload field as a FastHTML component.

        Args:
            *children: Optional children components.
            **kwargs: Additional keyword arguments.

        Returns:
            FT: FastHTML component for FileUpload.
        """
        error_cls = " govuk-file-upload--error" if self.error else ""
        return super().__ft__(
            fh.Div(
                fh.Input(
                    id=self.name,
                    name=self.name,
                    value=self.value,
                    type="file",
                    cls=f"govuk-file-upload{error_cls}",
                    aria_describedby=f"{self._id}-hint {self._id}-error",
                ),
                cls="govuk-drop-zone",
                data_module="govuk-file-upload",
            ),
            **self.kwargs,
        )


def _date_input_item(
    name: str, suffix: str, width: int = 2, value: str = "", error: bool = False
):
    """
    Date Input item e.g. Day, Month or Year used to produce underlying inputs
    in the composite field that is the :py:class:`DateInput` component.

    TODO: This should probably be a property in DateInput.

    Args:
        name (str): Name of the parent DateField.
        suffix (str): Suffix for this field e.g. "day", "month", "year".
        width (int, optional): Width of the field. Defaults to 2.
        value (str, optional): Value to assign to the input. Defaults to "".
        error (bool, optional): Error state. Defaults to False.

    Returns:
        FT: A FastHTML input component.
    """
    _id = f"{mkid(name)}-{suffix}"
    input_id = f"{_id}-input"
    label = Label(input_id, suffix.title(), extra_cls=" govuk-date-input__label")
    date_cls = "govuk-date-input__input"
    width_cls = f" govuk-input--width-{width}"
    error_cls = " govuk-input--error" if error else ""
    input = fh.Input(
        name=name,
        value=value,
        _id=input_id,
        _type="text",
        inputmode="numeric",
        cls=f"govuk-input {date_cls}{width_cls}{error_cls}",
    )
    return fh.Div(
        fh.Div(label, input, cls="govuk-form-group", _id=_id),
        cls="govuk-date-input__item",
    )


class DateInput(Field):
    """
    `GOV.UK Date input`_ component. Renders GDS-style composite field with day, month, year.

    You should use the date input component to help users enter a memorable date or one they
    can easily look up.

    Examples:

        >>> ds.DateInput("dob")
        # Renders the classic GOV.UK 3-input - Day, Month, Year - date field.

    See :py:class:`fast_gov_uk.design_system.contrib.PastDateInput` and
    :py:class:`fast_gov_uk.design_system.contrib.FutureDateInput` for fields
    that validate whether the date should be in the past or future respectively.

    Args:
        *args: Arguments for Field.
        **kwargs: Additional keyword arguments.

    .. _GOV.UK Date input: https://design-system.service.gov.uk/components/date-input/
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def value(self):
        return self._value

    @property
    async def clean(self):
        """
        Get "clean" data from a field when a form is being processed
        by a form Backend.
        """
        # Field not required
        if self.value == ["", "", ""]:
            return None
        try:
            day, month, year = self.value
            day, month, year = int(day), int(month), int(year)
            _date = date(day=day, month=month, year=year)
            return _date.isoformat()
        except ValueError:
            raise

    @value.setter
    def value(self, value):
        """
        Sets date value as well as error if the date value is invalid.
        """
        self._value = value or ("", "", "")
        day, month, year = self._value
        if self.required:
            if (not day or not month or not year):
                self.error = "This field is required."
                return
            try:
                day, month, year = int(day), int(month), int(year)
                _ = date(day=day, month=month, year=year)
            except (ValueError, TypeError):
                self.error = "Invalid values."

    @property
    def day_field(self):
        """
        Day field component.
        """
        return _date_input_item(
            self.name,
            "day",
            error=(self.error != ""),
            value=self.value[0] if self.value else "",
        )

    @property
    def month_field(self):
        """
        Month field component.
        """
        return _date_input_item(
            self.name,
            "month",
            error=(self.error != ""),
            value=self.value[1] if self.value else "",
        )

    @property
    def year_field(self):
        """
        Year field component.
        """
        return _date_input_item(
            self.name,
            "year",
            width=4,
            error=(self.error != ""),
            value=self.value[2] if self.value else "",
        )

    def __ft__(self, *children, **kwargs) -> fh.FT:
        """
        Renders the date input field as a FastHTML component.
        Args:
            *children: Optional children components.
            **kwargs: Additional keyword arguments.
        Returns:
            FT: FastHTML component for DateInput.
        """
        return super().__ft__(
            fh.Fieldset(
                fh.Div(
                    self.day_field,
                    self.month_field,
                    self.year_field,
                    cls="govuk-date-input",
                    _id=self._id,
                ),
                cls="govuk-fieldset",
                role="group",
                aria_describedby=f"{self._id}-hint",
            ),
            **self.kwargs,
        )


class ButtonStyle(Enum):
    """
    Enum to define GOV.UK supported button styles to be passed into
    :py:class:`Button` as `style` param.
    """
    PRIMARY = 1
    SECONDARY = 2
    WARNING = 3
    INVERSE = 4


def Button(
    text: str,
    style: ButtonStyle = ButtonStyle.PRIMARY,
    disabled: bool = False,
    prevent_double_click: bool = False,
    **kwargs,
) -> fh.FT:
    """
    `GOV.UK Button`_ component used to help users carry out an action like starting an
    application or saving their information.

    Write button text in sentence case, describing the action it performs. For example:

    - "Start now" at the start of your service
    - "Sign in" to an account a user has already created
    - "Continue" when the service does not save a user"s information
    - "Save and continue" when the service does save a user"s information
    - "Save and come back later" when a user can save their information and come back later
    - "Add another" to add another item to a list or group
    - "Pay" to make a payment
    - "Confirm and send" on a Check answers page that does not have any legal content a user must agree to
    - "Accept and send" on a Check answers page that has legal content a user must agree to
    - "Sign out" when a user is signed in to an account

    Examples:

        >>> ds.Button("Submit")
        # Renders a default button called "Submit"
        >>> ds.Button("Cancel", ButtonStyle.WARNING)
        # Renders a warning button
        >>> ds.Button("Cancel", disabled=True)
        # Renders a disabled button
        >>> ds.Button("Cancel", prevent_double_click=True)
        # Renders a button with some js that prevents double clicks - useful for payments

    Args:
        text (str): The text on the Button component.
        style (ButtonStyle, optional): The style of the Button component. Defaults to ButtonStyle.PRIMARY.
        disabled (bool, optional): Disable the button. Defaults to False.
        prevent_double_click (bool, optional): Prevent accidental double clicks. Defaults to False.
        **kwargs: Any extra args to pass to fh.Button.

    Returns:
        FT: A FastHTML Button component.

    .. _GOV.UK Button: https://design-system.service.gov.uk/components/button/
    """
    btn_cls = {
        ButtonStyle.PRIMARY: "govuk-button",
        ButtonStyle.SECONDARY: "govuk-button govuk-button--secondary",
        ButtonStyle.WARNING: "govuk-button govuk-button--warning",
        ButtonStyle.INVERSE: "govuk-button govuk-button--inverse",
    }
    return fh.Button(
        text,
        _type="submit",
        disabled=disabled,
        aria_disabled=disabled,
        data_prevent_double_click=prevent_double_click,
        data_module="govuk-button",
        cls=btn_cls[style],
        **kwargs,
    )


def StartButton(text: str, href: str, **kwargs) -> fh.FT:
    """
    `GOV.UK Start Button`_ component for call-to-actions. Does not submit form data.

    Examples:

        >>> ds.StartButton("Lets go")
        # Renders a start button with the label "Lets go"
        # This button acts more like a link - it results in a GET request instead of POST

    Args:
        text (str): Text on the Button component.
        href (str): URL of the target page.
        **kwargs: Additional keyword arguments.

    Returns:
        FT: A FastHTML StartButton component.

    .. _GOV.UK Start Button: https://design-system.service.gov.uk/components/button/
    """
    icon = fh.NotStr(
        '<svg class="govuk-button__start-icon" xmlns="http://www.w3.org/2000/svg" width="17.5" '
        'height="19" viewBox="0 0 33 40" aria-hidden="true" focusable="false">'
        '<path fill="currentColor" d="M0 0h13l20 20-20 20H0l20-20z" />'
        "</svg>"
    )
    return fh.A(
        text,
        icon,
        href=href,
        role="button",
        draggable="false",
        cls="govuk-button govuk-button--start",
        data_module="govuk-button",
        **kwargs,
    )


def ButtonGroup(*buttons: fh.FT, **kwargs) -> fh.FT:
    """
    `GOV.UK Button Group`_ component.

    Use a button group when two or more buttons are placed together.

    Examples:

        >>> continue = ds.Button("Continue")
        >>> cancel = ds.Button("Cancel")
        >>> ds.ButtonGroup(contiue, cancel)
        # Renders a button group

    Args:
        *buttons (FT): List of Button components.
        **kwargs: Additional keyword arguments.

    Returns:
        FT: A FastHTML component.

    .. _GOV.UK Button Group: https://design-system.service.gov.uk/components/button/
    """
    return fh.Div(
        *buttons,
        cls="govuk-button-group",
        **kwargs,
    )


def CookieBanner(
    service_name: str,
    *content: fh.FT,
    cookie_page_link: str = "/cookies",
    cookie_form_link: str = "/cookie-banner",
    confirmation: bool = False,
    **kwargs,
) -> fh.FT:
    """
    `GOV.UK Cookie Banner`_ component.

    Allow users to accept or reject cookies which are not essential to making your service work.

    This component needs links to 2 endpoints on the service -

        1. `cookie_page_link` is a link to the cookies page for your service.
        2. `cookie_form_link` is a link to the cookies form that processes the data when a user submits their cookie preferences through this component.

    Fast-gov-uk comes out of the box with default, minimal implementations of both. In fact,
    fast-gov-uk comes out of the box with a cookie banner for essentail cookies.

    In view of the above, you would only use this component if you are creating a custom cookie
    banner for your service.

    Examples:

        >>> ds.CookieBanner("Test")
        # Renders a standard cookie banner for essential cookies that works out of the box

    Args:
        service_name (str): Name of the service.
        *content (FT): Content of the CookieConfirmation component.
        cookie_page_link (str, optional): Link to the cookie settings page. Defaults to "/cookies".
        cookie_form_link (str, optional): Link to the cookie form submission page. Defaults to "/".
        confirmation (bool, optional): If True, the cookie confirmation is shown. Defaults to False.
        **kwargs: Additional keyword arguments.

    Returns:
        FT: A FastHTML CookieConfirmation component.

    .. _GOV.UK Cookie Banner: https://design-system.service.gov.uk/components/cookie-banner/
    """
    banner_buttons = ButtonGroup(
        Button("Accept additional cookies", value="yes", name="cookies[additional]"),
        Button("Reject additional cookies", value="no", name="cookies[additional]"),
        A("View cookies", href=cookie_page_link),
    )
    confirm_buttons = ButtonGroup(
        # TODO: Add some js to hide this when button is pressed
        Button("Hide cookie message", value="hide", name="cookies[additional]"),
    )
    button_group = confirm_buttons if confirmation else banner_buttons
    return fh.Div(
        fh.Div(
            fh.Div(
                fh.Div(
                    H2(f"Cookies for {service_name}"),
                    fh.Div(
                        *content,
                        cls="govuk-cookie-banner__content",
                    ),
                    cls="govuk-grid-column-two-thirds",
                ),
                cls="govuk-grid-row",
            ),
            fh.Form(
                button_group,
                hx_post=cookie_form_link,
                hx_target="#cookie-banner",
                hx_swap="outerHTML",
            ),
            cls="govuk-cookie-banner__message govuk-width-container",
        ),
        cls="govuk-cookie-banner",
        role="region",
        aria_label=f"Cookies on {service_name}",
        data_nosnippet=True,
        id="cookie-banner",
        **kwargs,
    )


class Fieldset(AbstractField):
    """
    `GOV.UK Fieldset`_ component used to group related form fields.

    Use the fieldset component when you need to show a relationship between multiple form inputs.
    For example, you may need to group a set of text inputs into a single fieldset when asking for
    an address in your service.

    Examples:

        >>> address1 = ds.TextInput("address1", label="Address line 1")
        >>> address2 = ds.TextInput("address2", label="Address line 2")
        >>> town = ds.TextInput("town", label="Town or city")
        >>> postcode = ds.TextInput("postcode", label="Postcode")
        >>> ds.Fieldset(address1, address2, town, postcode)
        # Renders an address fieldset containing multiple, related fields

    Args:
        *fields (Field): Fields to include in the fieldset.
        name (str, optional): Name of the fieldset. Defaults to "".
        legend (str, optional): The legend text for the fieldset.
        heading (str, optional): Heading size. Defaults to "l".
        **kwargs: Additional keyword arguments.

    .. _GOV.UK Fieldset: https://design-system.service.gov.uk/components/fieldset/
    """

    def __init__(self, *fields: Field, name: str = "", legend: str = "", heading: str = "l", **kwargs):
        self.fields = fields
        self.name = name
        self.legend = legend
        self.heading = heading
        self.kwargs = kwargs

    def __ft__(self):
        """
        Renders the fieldset as a FastHTML component.

        Returns:
            FT: A FastHTML Fieldset component.
        """
        heading_cls = f" govuk-fieldset__legend--{self.heading}" if self.heading else ""
        return fh.Fieldset(
            fh.Legend(
                fh.H1(
                    self.legend,
                    cls="govuk-fieldset__heading",
                ),
                cls=f"govuk-fieldset__legend{heading_cls}",
            ),
            *self.fields,
            cls="govuk-fieldset",
            **self.kwargs,
        )



from pathlib import Path

import fasthtml.common as fh
from notifications_python_client import notifications as notify

import fast_gov_uk.design_system as ds
from fast_gov_uk import forms


GOV_UK_HTTP_HEADERS = [
    fh.Link(rel="stylesheet", href="/govuk-frontend-5.11.1.min.css", type="text/css"),
    fh.Meta(name="theme-color", content="#1d70b8"),
    fh.Link(rel="icon", sizes="48x48", href="/assets/rebrand/images/favicon.ico"),
    fh.Link(
        rel="icon",
        sizes="any",
        href="/assets/rebrand/images/favicon.svg",
        type="image/svg+xml",
    ),
    fh.Link(
        rel="mask-icon",
        href="/assets/rebrand/images/govuk-icon-mask.svg",
        color="#1d70b8",
    ),
    fh.Link(rel="apple-touch-icon", href="/assets/rebrand/images/govuk-icon-180.png"),
]


def not_found(req, exc):
    return ds.Page(
        ds.H1("Page not found"),
        ds.P("If you typed the web address, check it is correct."),
        ds.P("If you pasted the web address, check you copied the entire address."),
        ds.P(
            "If the web address is correct or you selected a link or button, ",
            ds.A("contact us", "/forms/contact-us"),
            "to speak to someone and get help.",
        ),
    )


def problem_with_service(req, exc):
    return ds.Page(
        ds.H1("Sorry, there is a problem with this service"),
        ds.P("Try again later."),
        ds.P(
            ds.A("Contact us", "/forms/contact-us"),
            " if you need to speak to someone about this.",
        ),
    )


def assets(fname: str, ext: str):
    """Serve static assets from the assets directory."""
    assets = Path(__file__).parent / "assets"
    return fh.FileResponse(f"{assets}/{fname}.{ext}")


def footer():
    """
    Footer that will be rendered on every `Page` in the service.

    You can override this like so -

        @fast.page
        def footer():
            return ds.Footer(("My custom footer link", "/link"), ..)
    """
    return ds.Footer(
        ("Cookies", "/cookies"),
        ("Give feedback", "/forms/feedback"),
    )


def phase():
    """
    Returns a phase banner snippet that is inserted into every
    `Page` of the service using htmx.

    You can override this like so -

        @fast.page
        def phase():
            return ds.PhaseBanner(
                ds.Span("This is a slightly older service.")
                phase="Beta",
            )
    """
    return ds.PhaseBanner(
        ds.Span(
            "This is a new service. Help us improve it and ",
            ds.A("give your feedback.", href="/forms/feedback"),
        ),
        phase="Alpha",
    )


def cookies():
    """
    This returns the standard GDS cookies page. E.g. -
    https://design-system.service.gov.uk/patterns/cookies-page/full-page/

    It is preset to Essential cookies but if you are using more cookies,
    you can override this as follows -

        @fast.page
        def cookies():
            return ds.Cookies(
                ds.H2("Analytical cookies")
                ds.P("...")
                ds.Table(...)
            )
    """
    return ds.Cookies()


def feedback(data=None):
    """
    Feedback form is common and recommended in gov.uk services. This form
    is an attempt to reproduce -

    https://www.gov.uk/service-manual/service-assessments/get-feedback-page

    This forms logs the feedback. If you want to change this form or change
    the Backend so that you get an email when a user submits feedback -

        @fast.form
        def feedback(data=None):
            return forms.Form(
                "feedback",
                ...
                backends=[forms.EmailBackend(...)]
            )
    """
    # A DBForm gets saved to the database when its valid
    return forms.Form(
        "feedback",
        ds.H1("Give feedback on Fast GOV UK"),
        ds.H2("Satisfaction survey"),
        ds.Radios(
            name="satisfaction",
            label="Overall, how satisfied did you feel about Fast Gov UK?",
            choices={
                "very-satisfied": "Very Satisfied",
                "satisfied": "Satisfied",
                "neutral": "Neither satisfied not dissatisfied",
                "dissatisfied": "Dissatisfied",
                "very-dissatisfied": "Very dissatisfied",
            },
            heading="s",
        ),
        ds.CharacterCount(
            name="comments",
            label="How could we improve this service?",
            maxchars=1200,
            required=False,
            hint=(
                "Do not include any personal or financial information, "
                "for example your national insurance number."
            ),
            heading="s",
        ),
        backends=[forms.LogBackend()],
        success_url="/",
        data=data,
        cta="Send feedback",
    )


def notifications(session):
    notifications = session.get("notifications", [])
    banners = ds.Div(
        *[
            ds.Notification(
                n["content"],
                title=n["title"],
                success=(n["title"] == "Success")
            )
            for n in notifications
        ]
    )
    session.pop("notifications", None)
    return banners


class Fast(fh.FastHTML):
    def __init__(self, settings=None, *args, **kwargs):
        super().__init__(
            pico=False,
            hdrs=GOV_UK_HTTP_HEADERS,
            cls="govuk-template govuk-template--rebranded",
            # TODO: Why do I have to do this?
            style="margin: 0px;",
            exception_handlers={404: not_found, 500: problem_with_service},
            # Set up session cookie
            session_cookie="session_cookie",
            max_age=24*60*60,
        )
        settings = settings or {}
        # Service name
        self.service_name = settings.get("SERVICE_NAME", "Fast")
        # Set up Database
        db_url = settings.get("DATABASE_URL", ":memory:")
        self.db = fh.database(db_url)
        # Set up flag for whether we are in dev mode
        self.dev = settings.get("DEV_MODE", True)
        # Initialize form registry
        self.forms = {}
        # Initialise wizard registry
        self.wizards = {}
        # Set up routes
        self.route("/{fname:path}.{ext:static}")(assets)
        self.route("/forms/{name}", methods=["GET", "POST"])(self.process_form)
        self.route("/wizards/{name}/{step}", methods=["GET", "POST"])(
            self.process_wizard
        )
        self.route("/cookie-banner", methods=["GET", "POST"])(self.cookie_banner)
        self.route()(footer)
        self.route()(phase)
        self.route()(cookies)
        self.form()(feedback)
        self.route("/notifications")(notifications)
        # Initialise notify client
        notify_key = settings.get("NOTIFY_API_KEY", "")
        if notify_key:
            self.notify_client = notify.NotificationsAPIClient(notify_key)

    def notify(self, template_id: str, email: str):
        if not hasattr(self, "notify_client"):
            raise ValueError("NOTIFY_API_KEY not configured.")

        async def _notifier(**kwargs):
            kwargs["service_name"] = self.service_name
            return self.notify_client.send_email_notification(
                email_address=email,
                template_id=template_id,
                personalisation=kwargs,
            )

        return _notifier

    def page(self, url=None):
        def page_decorator(func):
            _url = url or f"/{func.__name__}"
            self.route(_url)(func)
            return func

        if callable(url):
            # Used as @app.page
            func = url
            url = None
            return page_decorator(func)
        # Used as @app.page("/some-url")
        return page_decorator

    def form(self, url=None):
        def form_decorator(func):
            _url = url or func.__name__
            self.forms[_url] = func
            return func

        if callable(url):
            # Used as @app.form
            func = url
            url = None
            return form_decorator(func)
        # Used as @app.form("/some-url")
        return form_decorator

    def wizard(self, url=None):
        def wizard_decorator(func):
            _url = url or func.__name__
            self.wizards[_url] = func
            return func

        if callable(url):
            # Used as @app.form
            func = url
            url = None
            return wizard_decorator(func)
        # Used as @app.question("/some-url")
        return wizard_decorator

    async def process_form(self, req, name: str, post: dict):
        mkform = self.forms.get(name, None)
        if not mkform:
            raise fh.HTTPException(status_code=404)
        # If GET, just return the form
        if req.method == "GET":
            return mkform()
        # If POST, fill the form
        form = mkform(post)
        # If valid, process
        if form.valid:
            return await form.process(req)
        # Else return with errors
        return form

    async def process_wizard(
        self, req, session: dict, name: str, step: str, post: dict
    ):
        try:
            mkwizard = self.wizards[name]
            _step = int(step or "0")
        except (KeyError, ValueError):
            raise fh.HTTPException(status_code=404)
        # If GET, just return the form
        if req.method == "GET":
            return mkwizard(step=_step)
        # If POST, fill the form
        wizard = mkwizard(step=_step, data=post)
        # If step valid
        if wizard.step_valid:
            return await wizard.next_step(req)
        # Else return with errors
        return wizard

    def cookie_banner(self, req, post: dict, cookie_policy: str = ""):
        # TODO: Use @fast.form decorator for this with a custom
        # cookie backend
        banner = ds.CookieBanner(
            self.service_name,
            ds.P("We use some essential cookies to make this service work."),
            cookie_form_link="/cookie-banner",
            # TODO: ATM, we only use essential cookies so no need
            # to show the banner with accept/reject options
            confirmation=True,
        )
        if req.method == "POST":
            val = post.get("cookies[additional]", None)
            hide = val == "hide"
            cookie_val = "hide" if hide else ""
            # Cookie should expire in a year
            cookie_age = 365*24*60*60
            cookie = fh.cookie("cookie_policy", cookie_val, max_age=cookie_age)
            return "" if hide else banner, cookie
        hide = "hide" in cookie_policy
        return "" if hide else banner

    def add_notification(self, session, content, success=False):
        notifications = session.get("notifications", [])
        notifications.append({"title": "Success" if success else "Important", "content": content})
        session["notifications"] = notifications



"""
GOV.UK Design System components for navigation elements such as back link, skip link, exit this page
as well as breadcrumbs, service navigation and pagination.
"""

import fasthtml.common as fh

from .utils import Next, Previous


def BackLink(href: str, text: str = "Back", inverse: bool = False, **kwargs) -> fh.FT:
    """
    The `GOV.UK back link`_ component to help users navigate to the previous page
    in a multi-page transaction.

    Examples:

        >>> backlink = ds.BackLink(href="/back", text="Go back")
        >>> str(backlink)
        '<a href="/back" class="govuk-back-link">Go back</a>'

    Args:
        href (str): Link to the previous page.
        text (str, optional): The text to display in the link. Defaults to "Back".
        inverse (bool, optional): If True, applies an inverse style. Defaults to False.
        **kwargs: Additional keyword arguments.

    Returns:
        FT: A FastHTML BackLink component.

    .. _GOV.UK back link: https://design-system.service.gov.uk/components/back-link/
    """
    inverse_cls = " govuk-back-link--inverse" if inverse else ""
    return fh.A(
        text,
        href=href,
        cls=f"govuk-back-link{inverse_cls}",
        **kwargs,
    )


def SkipLink(
    href: str,
    text: str = "Skip to main content",
    **kwargs,
) -> fh.FT:
    """
    `GOV.UK skip link`_ component to help keyboard-only users skip to the main content on a page.

    Examples:
        >>> skip_link = ds.SkipLink(href="#main")
        >>> str(skip_link)
        '<a href="#main" data-module="govuk-skip-link" class="govuk-skip-link">Skip to main content</a>'

    Args:
        href (str): On-page anchor (e.g. #main) for the main content.
        text (str, optional): The text to display in the link. Defaults to "Skip to main content".
        **kwargs: Additional keyword arguments.

    Returns:
        FT: A FastHTML SkipLink component.

    .. _GOV.UK skip link: https://design-system.service.gov.uk/components/skip-link/
    """
    return fh.A(
        text, href=href, cls="govuk-skip-link", data_module="govuk-skip-link", **kwargs
    )


def Breadcrumbs(
    *links: tuple[str, str],
    collapse_on_mobile: bool = False,
    **kwargs,
) -> fh.FT:
    """
    `GOV.UK breadcrumbs`_ component to help users understand where they are within a website's structure
    and move between levels.

    Examples:

        >>> ds.Breadcrumbs(
        ...     ("Home", "/"),
        ...     ("Section", "/section"),
        ... )
        # Home > Section

    Args:
        *links (tuple[str, str]): Text & URL for breadcrumb links.
        collapse_on_mobile (bool, optional): Make breadcrumbs responsive. Defaults to False.
        **kwargs: Additional keyword arguments.

    Returns:
        FT: A FastHTML Breadcrumbs component.

    .. _GOV.UK Breadcrumbs: https://design-system.service.gov.uk/components/breadcrumbs/
    """
    collapse_cls = (
        " govuk-breadcrumbs--collapse-on-mobile" if collapse_on_mobile else ""
    )
    return fh.Nav(
        fh.Ol(
            *[
                fh.Li(
                    fh.A(text, href=href, cls="govuk-breadcrumbs__link"),
                    cls="govuk-breadcrumbs__list-item",
                )
                for text, href in links
            ],
            cls="govuk-breadcrumbs__list",
        ),
        cls=f"govuk-breadcrumbs{collapse_cls}",
        aria_label="Breadcrumb",
        **kwargs,
    )


def ExitPage(
    text: str = "Exit this page",
    href: str = "https://www.bbc.co.uk/weather",
    **kwargs,
) -> fh.FT:
    """
    `GOV.UK exit page`_ component to give users a way to quickly and safely exit a service,
    website or application.

    Examples:

        >>> ds.ExitPage()
        # An Exit this page button linking to BBC weather service.
        >>> ds.ExitPage(text="Leave now", href="https://google.com")
        # An Exit this page button with the text "Leave now" linking to Google.

    Args:
        text (str, optional): The text to display on the ExitPage component. Defaults to "Exit this page".
        href (str, optional): The URL the link points to. Defaults to BBC weather service.
        **kwargs: Additional keyword arguments.

    Returns:
        FT: A FastHTML ExitPage component.

    .. _GOV.UK exit page: https://design-system.service.gov.uk/components/exit-this-page/
    """
    return fh.Div(
        fh.A(
            fh.Span(
                "Emergency",
                cls="govuk-visually-hidden",
            ),
            text,
            href=href,
            role="button",
            draggable="false",
            cls=(
                "govuk-button govuk-button--warning"
                " govuk-exit-this-page__button"
                " govuk-js-exit-this-page-button"
            ),
            data_module="govuk-button",
            rel="nofollow noreferrer",
        ),
        cls="govuk-exit-this-page",
        data_module="govuk-exit-this-page",
        **kwargs,
    )


def NavigationLink(
    text: str,
    href: str,
    active: bool = False,
    **kwargs,
) -> fh.FT:
    """
    Navigation link to pass into the :py:meth:`Navigation` component.

    Examples:

        >>> nav_link = ds.NavigationLink(text="Home", href="/")
        >>> str(nav_link)
        '<li class="govuk-service-navigation__item">'
            '<a href="/" class="govuk-service-navigation__link">home</a>'
        '</li>'

    Args:
        text (str): Text for the NavigationLink.
        href (str): Link for the NavigationLink.
        active (bool, optional): Is the NavigationLink active? Defaults to False.
        **kwargs: Additional keyword arguments.

    Returns:
        FT: A FastHTML NavigationLink component.
    """
    return fh.Li(
        fh.A(
            fh.Strong(text, cls="govuk-service-navigation__active-fallback")
            if active
            else text,
            href=href,
            cls="govuk-service-navigation__link",
            aria_current="true" if active else False,
        ),
        cls=(
            "govuk-service-navigation__item"
            f"{' govuk-service-navigation__item--active' if active else ''}"
        ),
        **kwargs,
    )


def Navigation(
    *links: fh.FT,
    service_name: str = "",
    **kwargs,
) -> fh.FT:
    """
    `GOV.UK Service Navigation`_ component to help users understand that they're using
    your service and lets them navigate around your service.

    You can pass it a list of :py:meth:`NavigationLink` components to build the
    navigation menu.

    Optionally, you can provide a `service_name` to display the name of your service
    on the navigation menu.

    Examples:

        >>> nav = ds.Navigation(
        ...     ds.NavigationLink(text="Home", href="/", active=True),
        ...     ds.NavigationLink(text="About", href="/about"),
        ...     service_name="My Service",
        ... )
        # *Home* | About

    Args:
        *links (FT): List of NavigationLink components.
        service_name (str, optional): Name of the service. Defaults to "".
        **kwargs: Additional keyword arguments.

    Returns:
        FT: A FastHTML Navigation component.

    .. _GOV.UK Service Navigation: https://design-system.service.gov.uk/components/service-navigation/
    """
    return fh.Section(
        fh.Div(
            fh.Div(
                fh.Span(
                    fh.A(service_name, href="/", cls="govuk-service-navigation__link"),
                    cls="govuk-service-navigation__service-name",
                )
                if service_name
                else "",
                fh.Nav(
                    fh.Button(
                        "Menu",
                        type="button",
                        cls="govuk-service-navigation__toggle govuk-js-service-navigation-toggle",
                        aria_controls="navigation",
                        hidden=True,
                    ),
                    fh.Ul(
                        *links,
                        cls="govuk-service-navigation__list",
                        id="navigation",
                    ),
                    aria_label="Menu",
                    cls="govuk-service-navigation__wrapper",
                ),
                cls="govuk-service-navigation__container",
            ),
            cls="govuk-width-container",
        ),
        cls="govuk-service-navigation",
        aria_label="Service information",
        data_module="govuk-service-navigation",
        **kwargs,
    )


def _pagination_prev(href: str) -> fh.FT:
    """
    Previous link for pagination that looks like a left arrow.

    Args:
        href (str): URL for the previous page.

    Returns:
        FT: A FastHTML previous page component.
    """
    return fh.Div(
        fh.A(
            fh.NotStr(Previous()),
            fh.Span(
                "Previous",
                fh.Span(" page", cls="govuk-visually-hidden"),
                cls="govuk-pagination__link-title",
            ),
            href=href,
            cls="govuk-link govuk-pagination__link",
            rel="prev",
        ),
        cls="govuk-pagination__prev",
    )


def _pagination_next(href: str) -> fh.FT:
    """
    Next link for pagination that looks like a right arrow.

    Args:
        href (str): URL for the next page.

    Returns:
        FT: A FastHTML next page component.
    """
    return fh.Div(
        fh.A(
            fh.Span(
                "Next",
                fh.Span(" page", cls="govuk-visually-hidden"),
                cls="govuk-pagination__link-title",
            ),
            fh.NotStr(Next()),
            href=href,
            cls="govuk-link govuk-pagination__link",
            rel="next",
        ),
        cls="govuk-pagination__next",
    )


def PaginationLink(
    label: str,
    href: str,
    active: bool = False,
    **kwargs,
) -> fh.FT:
    """
    Pagination link to pass into the :py:meth:`Pagination` component.

    Examples:

        >>> page_link = ds.PaginationLink(label="2", href="/page/2")
        >>> str(page_link)
        '<li class="govuk-pagination__item">'
            '<a href="/2" aria-label="Page 1" class="govuk-link govuk-pagination__link">1</a>'
        '</li>'

    Args:
        label (str): Label for the link.
        href (str): URL for the page.
        active (bool, optional): Is this the current page? Defaults to False.
        **kwargs: Additional keyword arguments.

    Returns:
        FT: A FastHTML component.
    """
    active_cls = " govuk-pagination__item--current" if active else ""
    return fh.Li(
        fh.A(
            label,
            href=href,
            cls="govuk-link govuk-pagination__link",
            aria_label=f"Page {label}",
            aria_current="page" if active else "",
        ),
        cls=f"govuk-pagination__item{active_cls}",
        **kwargs,
    )


def Pagination(
    *links: fh.FT,
    prev_link: str = "",
    next_link: str = "",
    **kwargs,
) -> fh.FT:
    """
    `GOV.UK Pagination`_ component to help users navigate forwards and backwards
    through a series of pages.

    You can pass it a list of :py:meth:`PaginationLink` components to build the
    Pagination.

    `prev_link` and `next_link` are optional links for previous and next pages.


    Examples:

        >>> pagination = ds.Pagination(
        ...     ds.PaginationLink(label="1", href="/page/1"),
        ...     ds.PaginationLink(label="2", href="/page/2", active=True),
        ...     ds.PaginationLink(label="3", href="/page/3"),
        ...     prev_link="/page/1",
        ...     next_link="/page/3",
        ... )
        # ← Previous 1 *2* 3 Next →

    Args:
        *links (FT): List of PaginationLink components.
        prev_link (str, optional): Link for previous page. Defaults to "".
        next_link (str, optional): Link for next page. Defaults to "".
        **kwargs: Additional keyword arguments.

    Returns:
        FT: A FastHTML Pagination component.

    .. _GOV.UK Pagination: https://design-system.service.gov.uk/components/pagination/
    """
    return fh.Nav(
        _pagination_prev(prev_link) if prev_link else "",
        fh.Ul(
            *links,
            cls="govuk-pagination__list",
        ),
        _pagination_next(next_link) if next_link else "",
        cls="govuk-pagination",
        aria_label="Pagination",
        **kwargs,
    )


def PaginationBlock(
    prev: tuple[str, str],
    next: tuple[str, str],
    **kwargs,
) -> fh.FT:
    """
    `GOV.UK block-style pagination`_ component to let users navigate through related content
    that has been split across multiple pages.

    Examples:

        >>> pagination = ds.PaginationBlock(
        ...     prev=("Previous page", "/previous"),
        ...     next=("Next page", "/next"),
        ... )
        # ← Previous: Previous page
        #   Next: Next page →

    Args:
        prev (tuple): Text and Link for previous page.
        next (tuple): Text and Link for next page.
        **kwargs: Additional keyword arguments.

    Returns:
        FT: A FastHTML Pagination component.

    .. _GOV.UK block-style pagination: https://design-system.service.gov.uk/components/pagination/
    """
    prev_label, prev_link = prev
    prev_component = fh.Div(
        fh.A(
            fh.NotStr(Previous()),
            fh.Span(
                "Previous",
                fh.Span(" page", cls="govuk-visually-hidden"),
                cls="govuk-pagination__link-title",
            ),
            fh.Span(":", cls="govuk-visually-hidden"),
            fh.Span(prev_label, cls="govuk-pagination__link-label"),
            href=prev_link,
            cls="govuk-link govuk-pagination__link",
            rel="prev",
        ),
        cls="govuk-pagination__prev",
    )
    next_label, next_link = next
    next_component = fh.Div(
        fh.A(
            fh.NotStr(Next()),
            fh.Span(
                "Next",
                fh.Span(" page", cls="govuk-visually-hidden"),
                cls="govuk-pagination__link-title",
            ),
            fh.Span(":", cls="govuk-visually-hidden"),
            fh.Span(next_label, cls="govuk-pagination__link-label"),
            href=next_link,
            cls="govuk-link govuk-pagination__link",
            rel="next",
        ),
        cls="govuk-pagination__next",
    )
    return fh.Nav(
        prev_component,
        next_component,
        cls="govuk-pagination govuk-pagination--block",
        aria_label="Pagination",
        **kwargs,
    )



"""
GOV.UK Design System components such as Inset, Detail, Panel, Tag, Warning,
Notification, Accordion, Tab, ErrorSummary, Table, TaskList, SummaryList,
and SummaryCard.

Each component is implemented as a function that returns a FastHTML
component with the appropriate GOV.UK classes and structure.
"""

import fasthtml.common as fh

from .utils import mkid


def Inset(text: str, **kwargs) -> fh.FT:
    """
    `GOV.UK Inset text`_ component. Use the inset text component to differentiate
    a block of text from the content that surrounds it, for example: quotes, examples
    and additional information about the page.

    Examples:

        >>> inset = ds.Inset("This is an inset text")
        # Renders inset text -
        # | This is an inset text

    Args:
        text (str): The main text to display.
        **kwargs: Additional keyword arguments.

    Returns:
        FT: A FastHTML Inset text component.

    .. _GOV.UK Inset text: https://design-system.service.gov.uk/components/inset-text/
    """
    return fh.Div(text, cls="govuk-inset-text", **kwargs)


def Detail(
    summary: str,
    *content: fh.FT | str,
    open: bool = False,
    **kwargs,
) -> fh.FT:
    """
    `GOV.UK Detail`_ component to make a page easier to scan by letting users reveal
    more detailed information only if they need it.

    Examples:

        >>> detail = ds.Detail("More details", "This is the detailed content.")
        # Renders detail component with summary and content.
        # ▼ More details
        # | This is the detailed content.

    Args:
        summary (str): The summary text for the detail.
        *content (FT or str): The content to display when the detail is expanded.
        open (bool, optional): If True, the detail is initially open. Defaults to False.
        **kwargs: Additional keyword arguments.

    Returns:
        FT: A FastHTML Detail component.

    .. _GOV.UK Detail: https://design-system.service.gov.uk/components/details/
    """
    return fh.Details(
        fh.Summary(
            fh.Span(summary, cls="govuk-details__summary-text"),
            cls="govuk-details__summary",
        ),
        fh.Div(
            *content,
            cls="govuk-details__text",
        ),
        cls="govuk-details",
        open=open,
        **kwargs,
    )


def Panel(
    *content: fh.FT | str,
    title: str = "",
    **kwargs,
) -> fh.FT:
    """
    `GOV.UK Panel`_ component - a visible container used on confirmation or
    results pages to highlight important content.

    Examples:

        >>> panel = ds.Panel("Panel content.", title="Test")
        # Renders panel component with title and content.
        # -----------------------
        # | Test
        # | Panel content.
        # -----------------------

    Args:
        *content (FT or str): The content to display in the panel.
        title (str, optional): The title of the panel. Defaults to "".
        **kwargs: Additional keyword arguments.

    Returns:
        FT: A FastHTML Panel component.

    .. _GOV.UK Panel: https://design-system.service.gov.uk/components/panel/
    """
    return fh.Div(
        fh.H1(title, cls="govuk-panel__title") if title else "",
        fh.Div(*content, cls="govuk-panel__body"),
        cls="govuk-panel govuk-panel--confirmation",
        **kwargs,
    )


def Tag(
    text: str,
    color: str = "",
    **kwargs,
) -> fh.FT:
    """
    `GOV.UK Tag`_ component used to show users the status of something.

    Examples:

        >>> tag = ds.Tag("In progress", color="blue")
        # Renders tag component with blue color.
        # [ In progress ]

    Args:
        text (str): The text to display in the tag.
        color (str, optional): The color of the tag. Defaults to "".
        **kwargs: Additional keyword arguments.

    Returns:
        FT: A FastHTML Tag component.

    .. _GOV.UK Tag: https://design-system.service.gov.uk/components/tag/
    """
    colors = {
        "": "",
        "blue": " govuk-tag--blue",
        "green": " govuk-tag--green",
        "red": " govuk-tag--red",
        "yellow": " govuk-tag--yellow",
        "grey": " govuk-tag--grey",
        "turquoise": " govuk-tag--turquoise",
        "light-blue": " govuk-tag--light-blue",
        "purple": " govuk-tag--purple",
        "pink": " govuk-tag--pink",
        "orange": " govuk-tag--orange",
    }
    return fh.Strong(text, cls=f"govuk-tag{colors.get(color)}", **kwargs)


def Warning(
    *content: fh.FT | str,
    **kwargs,
) -> fh.FT:
    """
    `GOV.UK Warning`_ text component used when you need to warn users about
    something important, such as legal consequences of an action, or lack
    of action, that they might take.

    Examples:

        >>> warning = ds.Warning("This is a warning message.")
        # Renders warning component with content.
        # ! This is a warning message.

    Args:
        *content (FT or str): The content to display in the warning.
        **kwargs: Additional keyword arguments.

    Returns:
        FT: A FastHTML Warning component.

    .. _GOV.UK Warning: https://design-system.service.gov.uk/components/warning-text/
    """
    return fh.Div(
        fh.Span(
            "!",
            cls="govuk-warning-text__icon",
            aria_hidden="true",
        ),
        fh.Strong(
            fh.Span("Warning", cls="govuk-visually-hidden"),
            *content,
            cls="govuk-warning-text__text",
        ),
        cls="govuk-warning-text",
        **kwargs,
    )


def NotificatonLink(
    text: str,
    href: str = "#",
    **kwargs,
) -> fh.FT:
    """
    NotificationLink component used in :py:meth:`Notification`.

    Examples:

        >>> link = ds.NotificatonLink("Click here", href="/more-info")
        >>> str(link)
        <a class="govuk-notification-banner__link" href="/more-info">Click here</a>

    Args:
        text (str): The text to display in the link.
        href (str, optional): The URL the link points to. Defaults to "#".
        **kwargs: Additional keyword arguments.

    Returns:
        FT: A FastHTML NotificationLink component.
    """
    cls = "govuk-notification-banner__link"
    return fh.A(text, href=href, cls=cls, **kwargs)


def NotificatonHeading(*content: fh.FT | str, **kwargs) -> fh.FT:
    """
    Notification heading component usef in :py:meth:`Notification`.

    Examples:

        >>> heading = ds.NotificatonHeading("Important update")
        >>> str(heading)
        <p class="govuk-notification-banner__heading">Important update</p>

    Args:
        *content (FT or str): The content to display in the heading.
        **kwargs: Additional keyword arguments.

    Returns:
        FT: A FastHTML H2 component.
    """
    return fh.P(
        *content,
        cls="govuk-notification-banner__heading",
        **kwargs,
    )


def Notification(
    *content: fh.FT,
    title: str = "Important",
    success: bool = False,
    **kwargs,
) -> fh.FT:
    """
    `GOV.UK Notification banner`_ component used to tell the user about something
    they need to know about, but that's not directly related to the page content.

    A notification banner lets you tell the user about something that's not directly
    relevant to the thing they're trying to do on that page of the service.

    For example:

        - telling the user about a problem that's affecting the service as a whole
        - telling the user about something that affects them in particular
        - telling the user about the outcome of something they've just done on a previous page

    Examples:

        >>> notification = ds.Notification(
        ...     ds.NotificatonHeading("Service update"),
        ...     "Our service will be down for maintenance on Saturday.",
        ...     title="Important",
        ... )
        # Renders notification banner with heading and content.
        # ------------------------------
        # | Important
        # | Service update
        # | Our service will be down for maintenance on Saturday.
        # ------------------------------

    Args:
        *content (FT): The content to display in the notification banner.
        title (str, optional): The title of the notification. Defaults to "Important".
        success (bool, optional): If True, applies a success style. Defaults to False.
        **kwargs: Additional keyword arguments.

    Returns:
        FT: A FastHTML Notification component.

    .. _GOV.UK Notification banner: https://design-system.service.gov.uk/components/notification-banner/
    """
    success_cls = " govuk-notification-banner--success" if success else ""
    return fh.Div(
        fh.Div(
            fh.H2(
                title,
                cls="govuk-notification-banner__title",
                id="govuk-notification-banner-title",
            ),
            cls="govuk-notification-banner__header",
        ),
        fh.Div(content, cls="govuk-notification-banner__content"),
        cls=f"govuk-notification-banner{success_cls}",
        role="alert",
        aria_labelledby="govuk-notification-banner-title",
        data_module="govuk-notification-banner",
        **kwargs,
    )


def _accordion_section(accordion_id, n, heading, summary, content, open=False):
    """
    Helper function to create an accordion section for :py:meth:`Accordion`.

    Args:
        accordion_id (str): Id for accordion.
        n (int): The section number.
        heading (str): The heading of the section.
        summary (str): The summary of the section.
        content (str): The content of the section.
        open (bool, optional): If True, the section is initially open. Defaults to False.

    Returns:
        FT: A FastHTML accordion section component.
    """
    return fh.Div(
        fh.Div(
            fh.H2(
                fh.Span(
                    heading,
                    cls="govuk-accordion__section-button",
                    id=f"{accordion_id}-heading-{n}",
                ),
                cls="govuk-accordion__section-heading",
            ),
            fh.Div(
                summary,
                cls="govuk-accordion__section-summary govuk-body",
                id=f"{accordion_id}-summary-{n}",
            )
            if summary
            else "",
            cls="govuk-accordion__section-header",
        ),
        fh.Div(
            content,
            cls="govuk-accordion__section-content",
            id=f"{accordion_id}-content-{n}",
        ),
        cls="govuk-accordion__section",
        open=open,
    )


def Accordion(*sections: dict, accordion_id="accordion", **kwargs) -> fh.FT:
    """
    `GOV.UK Accordion`_ component - lets users show and hide sections of
    related content on a page.

    Only use an accordion if there's evidence it's helpful for the user to:

        - see an overview of multiple, related sections of content
        - choose to show and hide sections that are relevant to them
        - look across information that might otherwise be on different pages

    Examples:

        >>> accordion = ds.Accordion(
        ...     {"heading": "Section 1", "content": "Content 1.", "open": True},
        ...     {"heading": "Section 2", "summary": "Summary 2.", "content": "Content 2."},
        ... )
        # Renders accordion with two sections, second section is open by default.
        # ------------------------------
        # | Section 1
        # | ▲ Hide
        # | Content for section 1.
        # ------------------------------
        # | Section 2
        # | Summary for section 2.
        # | ▼ Show
        # ------------------------------

    Args:
        *sections (dict): Sections to include in the accordion.
        accordion_id (str, optional): Id for accordion. Defaults to "accordion".
        **kwargs: Additional keyword arguments.

    Returns:
        FT: A FastHTML Accordion component.

    .. _GOV.UK Accordion: https://design-system.service.gov.uk/components/accordion/
    """
    return fh.Div(
        *[
            _accordion_section(
                accordion_id,
                n + 1,
                section["heading"],
                section.get("summary", ""),
                section["content"],
                open=section.get("open", False),
            )
            for n, section in enumerate(sections)
        ],
        cls="govuk-accordion",
        data_module="govuk-accordion",
        **kwargs,
    )


def _tab_panel(heading, content, active=False):
    """
    Helper function to create a tab panel for :py:meth:`Tab`.

    Args:
        heading (str): The heading of the panel.
        content (str): The content of the panel.
        active (bool, optional): Tab is active. Defaults to False.

    Returns:
        FT: A FastHTML tab section component.
    """
    tab_id = mkid(heading)
    active_cls = "" if active else " govuk-tabs__panel--hidden"
    return fh.Div(
        fh.H2(heading, cls="govuk-heading-l"),
        content,
        cls=f"govuk-tabs__panel{active_cls}",
        id=f"{tab_id}",
    )


def _tab_li(heading, active=False):
    """
    Helper function to create a tab list item for :py:meth:`Tab`.

    Args:
        heading (str): The heading of the tab.
        active (bool, optional): Tab is active. Defaults to False.

    Returns:
        FT: A FastHTML tab list item component.
    """
    tab_id = mkid(heading)
    active_cls = " govuk-tabs__list-item--selected" if active else ""
    return fh.Li(
        fh.A(
            heading,
            href=f"#{tab_id}",
            cls="govuk-tabs__tab",
        ),
        cls=f"govuk-tabs__list-item{active_cls}",
    )


def Tab(*panels: dict, title="", **kwargs) -> fh.FT:
    """
    `GOV.UK Tab`_ componen - lets users navigate between related sections
    of content, displaying one section at a time.

    Tabs can be a helpful way of letting users quickly switch between related
    information if:

        - your content can be usefully separated into clearly labelled sections
        - the first section is more relevant than the others for most users
        - users will not need to view all the sections at once

    Examples:

        >>> tab = ds.Tab(
        ...     {"heading": "Tab 1", "content": "Content 1."},
        ...     {"heading": "Tab 2", "content": "Content 2."},
        ...     title="Example Tabs",
        ... )
        # Renders tab component with two panels.
        # ------------------------------
        # | Example Tabs
        # | [Tab 1] Tab 2
        # | Content 1.
        # ------------------------------

    Args:
        *panels (dict): Panels to include in the tab.
        title (str, optional): Title of the tab. Defaults to "".
        **kwargs: Additional keyword arguments.

    Returns:
        FT: A FastHTML Tab component.

    .. _GOV.UK Tab: https://design-system.service.gov.uk/components/tabs/
    """
    return fh.Div(
        fh.H2(title, cls="govuk-tabs__title"),
        fh.Ul(
            *[
                _tab_li(panel["heading"], active=(n == 0))
                for n, panel in enumerate(panels)
            ],
            cls="govuk-tabs__list",
        ),
        *[
            _tab_panel(panel["heading"], panel["content"], active=(n == 0))
            for n, panel in enumerate(panels)
        ],
        cls="govuk-tabs",
        data_module="govuk-tabs",
        **kwargs,
    )


def ErrorSummary(title: str, *links: fh.FT, **kwargs) -> fh.FT:
    """
    `GOV.UK Error Summary`_ component used at the top of a page to summarise
    any errors a user has made.

    Examples:

        >>> error_summary = ds.ErrorSummary(
        ...     "There is a problem",
        ...     ds.A("Error 1", href="#error1"),
        ...     ds.A("Error 2", href="#error2"),
        ... )
        # Renders error summary component with title and links.
        # ------------------------------
        # | There is a problem
        # | - Error 1
        # | - Error 2
        # ------------------------------

    Args:
        title (str): The title of the ErrorSummary component.
        *links (A): Links to include in the ErrorSummary component.
        **kwargs: Additional keyword arguments.

    Returns:
        FT: A FastHTML ErrorSummary component.

    .. _GOV.UK Error Summary: https://design-system.service.gov.uk/components/error-summary/
    """
    return fh.Div(
        fh.Div(
            fh.H2(
                title,
                cls="govuk-error-summary__title",
            ),
            fh.Div(
                fh.Ul(
                    *[fh.Li(link) for link in links],
                    cls="govuk-list govuk-error-summary__list",
                ),
                cls="govuk-error-summary__body",
            ),
            role="alert",
        ),
        cls="govuk-error-summary",
        data_module="govuk-error-summary",
        **kwargs,
    )


def _table_head(headers, numeric_cols, col_width):
    """
    Helper function to create a table head for :py:meth:`Table`.
    """
    ths = []
    for header in headers:
        cls = "govuk-table__header"
        if header in col_width:
            width_cls = col_width.get(header, "")
            cls += f" govuk-!-width-{width_cls}"
        if header in numeric_cols:
            cls += " govuk-table__header--numeric"
        th = fh.Th(header, scope="col", cls=cls)
        ths.append(th)
    return fh.Thead(
        fh.Tr(*ths, cls="govuk-table__row"),
        cls="govuk-table__head",
    )


def _table_body(data, header_cols, numeric_cols):
    """
    Helper function to create a table body for :py:meth:`Table`.
    """
    trs = []
    for row in data:
        tds = []
        for col, val in row.items():
            if col in header_cols:
                cls = "govuk-table__header"
                if col in numeric_cols:
                    cls += " govuk-table__header--numeric"
                td = fh.Th(val, cls=cls, scope="row")
            else:
                cls = "govuk-table__cell"
                if col in numeric_cols:
                    cls += " govuk-table__cell--numeric"
                td = fh.Td(val, cls=cls)
            tds.append(td)
        trs.append(fh.Tr(*tds, cls="govuk-table__row"))
    return fh.Tbody(*trs, cls="govuk-table__body")


def Table(
    data: list[dict],
    caption: str = "",
    header_cols: list | None = None,
    numeric_cols: list | None = None,
    col_width: dict | None = None,
    small_text: bool = False,
    **kwargs,
) -> fh.FT:
    """
    `GOV.UK Table`_ component - use this component to make information easier to compare
    and scan for users.

    Examples:

        >>> table = ds.Table(
        ...     data=[
        ...         {"Name": "Alice", "Age": 30, "City": "London"},
        ...         {"Name": "Bob", "Age": 25, "City": "Manchester"},
        ...     ],
        ...     caption="User Information",
        ...     header_cols=["Name"],
        ...     numeric_cols=["Age"],
        ...     col_width={"Name": "one-third", "Age": "one-sixth", "City": "half"},
        ...     small_text=True,
        ... )
        # Renders table component with caption, header columns, numeric columns, column widths, and small text.
        # ----------------------------------------------
        # | User Information
        # | Name       |  Age  |      City
        # | ----------------------------------------------
        # | Alice      |   30  |     London
        # | Bob        |   25  |  Manchester
        # ----------------------------------------------

    Args:
        data (list[dict]): Data for the Table component.
        caption (str, optional): The caption of the Table component. Defaults to "".
        header_cols (list, optional): List of columns that should be headers in each row. Defaults to None.
        numeric_cols (list, optional): List of columns that are numeric. Defaults to None.
        col_width (dict, optional): Override column widths. Defaults to None.
        small_text (bool, optional): Render a more compact table. Defaults to False.
        **kwargs: Additional keyword arguments.

    Returns:
        FT: A FastHTML Table component.

    .. _GOV.UK Table: https://design-system.service.gov.uk/components/table/
    """
    header_cols = header_cols or []
    numeric_cols = numeric_cols or []
    col_width = col_width or {}
    small_text_cls = " govuk-table--small-text-until-tablet" if small_text else ""
    _caption = fh.Caption(caption, cls="govuk-table__caption govuk-table__caption--m")
    return fh.Table(
        _caption if caption else "",
        _table_head(data[0].keys(), numeric_cols, col_width),
        _table_body(data, header_cols, numeric_cols),
        cls=f"govuk-table{small_text_cls}",
        **kwargs,
    )


def Task(
    label: str,
    href: str,
    completed: bool = False,
    hint: str = "",
    **kwargs,
) -> fh.FT:
    """
    Task component used in :py:meth:`TaskList`.

    Args:
        label (str): Label for the Task item.
        href (str): Link for the Task item.
        completed (bool, optional): Is the task completed? Defaults to False.
        hint (str, optional): Hint for the Task item. Defaults to "".
        **kwargs: Additional keyword arguments.

    Returns:
        FT: A FastHTML Task component.
    """
    _id = mkid(label)
    status_id = f"{_id}-status"
    hint_id = f"{_id}-hint"
    aria_hint_id = f"{hint_id} " if hint else ""
    return fh.Li(
        fh.Div(
            fh.A(
                label,
                href=href,
                cls="govuk-link govuk-task-list__link",
                aria_describedby=f"{aria_hint_id}{status_id}",
            ),
            fh.Div(
                hint,
                cls="govuk-task-list__hint",
                _id=hint_id,
            )
            if hint
            else "",
            cls="govuk-task-list__name-and-hint",
        ),
        fh.Div(
            "Completed",
            cls="govuk-task-list__status",
            _id=status_id,
        )
        if completed
        else fh.Div(
            fh.Strong(
                "Incomplete",
                cls="govuk-tag govuk-tag--blue",
            ),
            cls="govuk-task-list__status",
            _id=status_id,
        ),
        cls="govuk-task-list__item govuk-task-list__item--with-link",
        **kwargs,
    )


def TaskList(
    *tasks: fh.FT,
    **kwargs,
) -> fh.FT:
    """
    `GOV.UK TaskList`_ component.

    The task list component displays all the tasks a user needs to do, and allows users
    to easily identify which ones are done and which they still need to do.

    Use the task list if there's evidence that users:

        - do not want to, or cannot, complete all the tasks in one sitting
        - need to be able to choose the order they complete the tasks in

    Examples:

        >>> task_list = ds.TaskList(
        ...     ds.Task("Task 1", href="/task-1", completed=True),
        ...     ds.Task("Task 2", href="/task-2", completed=False, hint="This is a hint."),
        ... )
        # Renders task list component with two tasks.
        # ------------------------------
        # | Task 1            | Completed
        # | Task 2            | Incomplete
        # | This is a hint.
        # ------------------------------

    Args:
        *tasks (FT): Tasks for the TaskList.
        **kwargs: Additional keyword arguments.

    Returns:
        FT: A FastHTML TaskList component.

    .. _GOV.UK TaskList: https://design-system.service.gov.uk/components/task-list/
    """
    return fh.Ul(
        *[task for task in tasks],
        cls="govuk-task-list",
        **kwargs,
    )


def SummaryItem(key: str, value: str | fh.FT, *actions: fh.FT, **kwargs):
    """
    SummaryRow component - a list of these goes in to form a :py:meth:`SummaryList`.

    Args:
        key (str): Key for the SummaryRow.
        value (str | fh.FT): Content of the SummaryRow.
        *actions (fh.FT): Action(s) assigned to the SummaryRow.
        **kwargs: Additional keyword arguments.

    Returns:
        FT: A FastHTML SummaryRow component.
    """
    for action in actions:
        action.children = (
            *action.children,
            fh.Span(key.lower(), cls="govuk-visually-hidden"),
        )

    if not actions:
        actions_component = ""
    else:
        if len(actions) == 1:
            actions_component = fh.Dd(*actions, cls="govuk-summary-list__actions")
        else:
            actions_component = fh.Dd(
                fh.Ul(
                    *[
                        fh.Li(action, cls="govuk-summary-list__actions-list-item")
                        for action in actions
                    ],
                    cls="govuk-summary-list__actions-list",
                ),
                cls="govuk-summary-list__actions",
            )
    no_actions_cls = "" if actions else " govuk-summary-list__row--no-actions"

    return fh.Div(
        fh.Dt(key, cls="govuk-summary-list__key"),
        fh.Dd(value, cls="govuk-summary-list__value"),
        actions_component,
        cls=f"govuk-summary-list__row{no_actions_cls}",
        **kwargs,
    )


def SummaryList(*items: fh.FT, border: bool = True, **kwargs) -> fh.FT:
    """
    `GOV.UK Summary List`_ component.

    Use a summary list to summarise information, for example, a user's responses at
    the end of a form.

    :py:meth:`SummaryItem` represents a single row in the summary list, and
    :py:meth:`SummaryCard` is a higher-level component that uses :py:meth:`SummaryList`.

    Examples:

        >>> summary_list = ds.SummaryList(
        ...     ds.SummaryItem("Name", "Alice", ds.A("Change", href="/change-name")),
        ...     ds.SummaryItem("Age", "30", ds.A("Change", href="/change-age")),
        ... )
        # Renders summary list component with two items.
        # ------------------------------
        # | Name       | Alice        | Change
        # | Age        | 30           | Change
        # ------------------------------

    Args:
        *items (FT): List of SummaryItems.
        border (bool, optional): Choose if a border should be drawn. Defaults to True.
        **kwargs: Additional keyword arguments.

    Returns:
        FT: A FastHTML SummaryList component.

    .. _GOV.UK Summary List: https://design-system.service.gov.uk/components/summary-list/
    """
    no_border_cls = "" if border else " govuk-summary-list--no-border"
    return fh.Dl(
        *items,
        cls=f"govuk-summary-list{no_border_cls}",
        **kwargs,
    )


def SummaryCard(
    title: str, summary_list: fh.FT, actions: list[fh.FT] | None = None, **kwargs
) -> fh.FT:
    """
    `GOV.UK Summary Card`_ component.

    If you're showing multiple summary lists on a page, you can show each list within a
    summary card. This lets you visually separate each summary list and give each a title
    and some actions.

    Examples:

        >>> summary_card = ds.SummaryCard(
        ...     title="Your details",
        ...     summary_list=ds.SummaryList(
        ...         ds.SummaryItem("Name", "Alice", ds.A("Change", href="/change-name")),
        ...         ds.SummaryItem("Age", "30", ds.A("Change", href="/change-age")),
        ...     ),
        ...     actions=[ds.A("Delete", href="/delete")],
        ... )
        # Renders summary card component with title, summary list, and actions.
        # ------------------------------
        # | Your details         | Delete
        # | Name       | Alice        | Change
        # | Age        | 30           | Change
        # ------------------------------

    Args:
        title (str): Title of the card.
        summary_list (fh.FT): SummaryList component.
        actions (list[fh.FT], optional): List of card actions. Defaults to None.
        **kwargs: Additional keyword arguments.

    Returns:
        FT: A FastHTML component.

    .. _GOV.UK Summary Card: https://design-system.service.gov.uk/components/summary-list/
    """
    actions = actions or []

    for action in actions:
        action.children = (
            *action.children,
            fh.Span(f"({title})", cls="govuk-visually-hidden"),
        )

    actions_component = (
        fh.Ul(
            *[fh.Li(action, cls="govuk-summary-card__action") for action in actions],
            cls="govuk-summary-card__actions",
        )
        if actions
        else ""
    )

    return fh.Div(
        fh.Div(
            fh.H2(title, cls="govuk-summary-card__title"),
            actions_component,
            cls="govuk-summary-card__title-wrapper",
        ),
        fh.Div(
            summary_list,
            cls="govuk-summary-card__content",
        ),
        cls="govuk-summary-card",
        **kwargs,
    )
